"""
Basic image registration, generally best suited most suited for coherent image
collection. This is based pretty directly on an approach developed at Sandia and
generally referred to by the name "regi".

The relevant matlab code appears to be authored by Terry M. Calloway,
Sandia National Laboratories, and modified by Wade Schwartzkopf, NGA.
"""

__classification__ = 'UNCLASSIFIED'
__author__ = ["Thomas McCullough", "Terry M. Calloway", "Wade Schwartzkopf"]


import logging
from typing import List, Tuple, Dict, Union, Optional

import numpy
from scipy.signal import correlate2d
from scipy.interpolate import LinearNDInterpolator

from sarpy.io.general.base import BaseReader


logger = logging.getLogger(__name__)


def _validate_match_parameters(
        reference_size: Tuple[int, int],
        moving_size: Tuple[int, int],
        match_box_size: Tuple[int, int],
        moving_deviation: Tuple[int, int],
        decimation: Tuple[int, int]) -> None:
    """
    Validate the match paramaters based the size of the images.

    Parameters
    ----------
    reference_size : Tuple[int, int]
    moving_size : Tuple[int, int]
    match_box_size : Tuple[int, int]
    moving_deviation : Tuple[int, int]
    decimation : Tuple[int, int]
    """

    if not ((match_box_size[0] % 2) == 1 and match_box_size[0] > 1 and
            (match_box_size[1] % 2) == 1 and match_box_size[1] > 1):
        raise ValueError('The match box size must have both odd entries greater than 1')

    if not ((moving_deviation[0] % 2) == 1 and moving_deviation[0] > 1 and
            (moving_deviation[1] % 2) == 1 and moving_deviation[1] > 1):
        raise ValueError('The match box size must have both odd entries greater than 1')

    limit_fraction = 0.5
    if match_box_size[0]*decimation[0] > limit_fraction*reference_size[0] or \
            match_box_size[1]*decimation[1] > limit_fraction*reference_size[1]:
        raise ValueError(
            'The size of the match box - {} with decimation - {} is too large\n\t'
            'with respect tothe size of the reference image - {}'.format(
                match_box_size, decimation, reference_size))
    if match_box_size[0]*decimation[0] > limit_fraction*moving_size[0] or \
            match_box_size[1]*decimation[1] > limit_fraction*moving_size[1]:
        raise ValueError(
            'The size of the match box - {} with decimation - {} is too close\n\t'
            'to the size of the moving image - {}'.format(
                match_box_size, decimation, moving_size))


def _populate_difference_structure(
        mapping_values: List[List[Dict]]) -> None:
    """
    Helper function for populating derivative estimates into our structure.

    Parameters
    ----------
    mapping_values: List[List[dict]]
    """

    # NB: this assumes the expected structure

    def do_diff(the_diff, the_count, direction, ref_loc0, mov_loc0, ref_loc1, mov_loc1):
        if mov_loc0 is None or mov_loc1 is None:
            return the_diff, the_count
        the_diff += float(mov_loc1[direction] - mov_loc0[direction]) / \
                    float(ref_loc1[direction] - ref_loc0[direction])
        the_count += 1
        return the_diff, the_count

    def basic_estimate_diff(entry, i, j):
        ref_loc = entry['reference_location']
        mov_loc = entry['moving_location']
        if mov_loc is None:
            return

        if entry.get('row_derivative', None) is None:
            # calculate row derivative
            r_diff = 0.0
            r_count = 0
            # get value based on before
            if i > 0:
                o_entry = mapping_values[i-1][j]
                do_diff(r_diff, r_count, 0, ref_loc, mov_loc,
                        o_entry['reference_location'], o_entry['moving_location'])
            # get value based on after
            if i < len(mapping_values) - 1:
                o_entry = mapping_values[i+1][j]
                do_diff(r_diff, r_count, 0, ref_loc, mov_loc,
                        o_entry['reference_location'], o_entry['moving_location'])
            if r_count > 0:
                row_der = r_diff/float(r_count)
                entry['row_derivative'] = row_der
                if row_der < 0.0:
                    logger.warning('Entry ({}, {}) has negative row derivative ({})'.format(i, j, row_der))

        if entry.get('column_derivative', None) is None:
            # calculate the column derivative
            c_diff = 0.0
            c_count = 0
            # get the value based on before
            if j > 0:
                o_entry = mapping_values[i][j-1]
                do_diff(c_diff, c_count, 1, ref_loc, mov_loc,
                        o_entry['reference_location'], o_entry['moving_location'])
            # get value based on after
            if j < len(mapping_values[0]) - 1:
                o_entry = mapping_values[i][j+1]
                do_diff(c_diff, c_count, 1, ref_loc, mov_loc,
                        o_entry['reference_location'], o_entry['moving_location'])
            if c_count > 0:
                col_der = c_diff/float(c_count)
                entry['column_derivative'] = col_der
                if col_der < 0.0:
                    logger.warning('Entry ({}, {}) has negative column derivative ({})'.format(i, j, col_der))

    for row_index, grid_row in enumerate(mapping_values):
        for col_index, element in enumerate(grid_row):
            basic_estimate_diff(element, row_index, col_index)


def _subpixel_shift(values: numpy.ndarray) -> float:
    """
    This is simplified port of the SAR toolbox matlab function fin_minms. This
    uses data from an empirical fit derived from unknown origins to estimate where
    the "real" minimum occurred.

    Parameters
    ----------
    values : numpy.ndarray
        Must have length 3, with either `values[1] <= min(values[0], values[2])`
        (a minimization problem), or `values[1] >= max(values[0], values[2])`
        (a maximization problem). Maximization problems will be re-cast as
        minimization through inversion.

    Returns
    -------
    shift : float
        This values will be (-1, 1), with -1 corresponding to the first location,
        0 corresponding to the center location, and 1 corresponding to the final
        location.
    """

    if not (isinstance(values, numpy.ndarray) and values.ndim == 1 and values.size == 3):
        raise ValueError('The input must be a 1-d array with three entries.')

    if not (values[1] >= max(values[0], values[2]) or values[1] <= min(values[0], values[2])):
        raise ValueError(
            'The central entry must either be larger than the other two (maximization problem)\n\t'
            'or smaller than the other two (minimization problem) - values {}'.format(values))

    if values[1] >= max(values[0], values[2]):
        # recast maximization problem as minimization
        return _subpixel_shift(-values)

    if values[0] == values[1] or values[1] == values[2]:
        # no real information
        return 0.0
    if values[0] == values[2]:
        # it's symmetric
        return 0.0

    values = (values + numpy.min(values))  # ensure that everything is positive by shifting up

    # Here are the verbatim matlab comments:
    # Algorithm -
    #   xm = min
    #   r = (rmsmid-rmsmin)/(rmsmax-rmsmin)
    #   empirical fit (r,xm) resembles arc of circle centered at xc=-11/8
    #   xm = 2(xc-3/8) + sqrt(4(xc-3/8)**2 - (r-1)((r-1)-2(xc-1)))
    #   empirical fit (r,xm) resembles arc of circle centered at xc=-6/8
    #   xm = 1-xc - sqrt(0.125 + 2(0.25-xc)**2 - (x-xc)**2)

    nsr = 0.5
    noise = nsr*numpy.max(values)
    rms = numpy.sqrt(values - noise)
    rmsmin = rms[1]
    rmsmid = min(rms[0], rms[2])
    rmsmax = max(rms[0], rms[2])
    r = (rmsmid - rmsmin)/(rmsmax - rmsmin)
    rm1 = r - 1.
    fit_val = 12.25 - rm1*(rm1 + 4.75)
    if not (2.5*2.5 < fit_val < 4.5*4.5):
        # probaly impossible...
        return 0.0
    shift = -3.5 + numpy.sqrt(fit_val)
    return -shift if rms[0] < rms[2] else shift


def _max_correlation_step(
        reference_array: numpy.ndarray,
        moving_array: numpy.ndarray,
        do_subpixel: bool = False) -> Tuple[Optional[numpy.ndarray], Optional[float]]:
    """
    Find the best match location of the moving array inside the reference array.

    Parameters
    ----------
    reference_array : numpy.ndarray
    moving_array : numpy.ndarray
    do_subpixel : bool
        Include a subpixel registration effort?

    Returns
    -------
    best_location : None|numpy.ndarray
        Will return `None` if there is no information, i.e. the reference patch
        or moving patch is all 0. Otherwise, this will be a numpy array
        `[row, column]` of the location of highest correlation, determined via
        :func:`numpy.argmax`.
    maximum_correlation : None|float
    """

    if reference_array.ndim != 2 or moving_array != 2:
        raise ValueError('Input arrays must be 2-dimensional')
    if reference_array.shape[0] < moving_array.shape[0] or reference_array.shape[1] < moving_array.shape[1]:
        raise ValueError(
            'It is required that the moving array (shape {}) is strictly contained\n\t'
            'inside  the reference array (shape {})'.format(moving_array.shape, reference_array.shape))

    # NB: sqrt suggested by matlab, presumably to dampen the importance of bright returns?
    reference_array = numpy.sqrt(numpy.abs(reference_array))
    if numpy.all(reference_array == 0):
        return None, None

    moving_array = numpy.sqrt(numpy.abs(moving_array))
    if numpy.all(moving_array == 0):
        return None, None

    kernel = numpy.ones(reference_array.shape, dtype='float32')

    # now, find the best match
    match_values = correlate2d(reference_array, moving_array, mode='valid')
    norm_values = correlate2d(kernel, moving_array*moving_array, mode='valid')
    mask = (norm_values > 0)

    # reduce to dot product of the unit vectors, and we pick the best match
    match_values[mask] /= numpy.sqrt(norm_values[mask])

    # raw maximum location
    raw_max_location = numpy.unravel_index(numpy.argmax(match_values), match_values.shape)
    maximum_value = match_values[raw_max_location[0], raw_max_location[1]]

    if do_subpixel:
        sub_shift = numpy.zeros((2,), dtype='float64')
        sub_shift[0] += _subpixel_shift(match_values[raw_max_location[0]-1:raw_max_location[0]+1, raw_max_location[1]])
        sub_shift[1] += _subpixel_shift(match_values[raw_max_location[0], raw_max_location[1]-1:raw_max_location[1]+1])
        raw_max_location += sub_shift

    if (moving_array.shape[0] % 2) == 0 or (moving_array.shape[1] % 2) == 0:
        shift = numpy.zeros((2, ), dtype='float64')
    else:
        shift = numpy.zeros((2, ), dtype='int64')
    shift[0] = 0.5*(moving_array.shape[0] - 1)
    shift[1] = 0.5*(moving_array.shape[1] - 1)
    return shift + raw_max_location, maximum_value


def _single_step_location(
        reference_data: Union[BaseReader, numpy.ndarray],
        reference_index: Optional[int],
        reference_size: Tuple[int, int],
        moving_data: Union[BaseReader, numpy.ndarray],
        moving_index: Optional[int],
        moving_size: Tuple[int, int],
        reference_location: Tuple[int, int],
        moving_location: Tuple[int, int],
        match_box_size: Tuple[int, int] = (25, 25),
        moving_deviation: Tuple[int, int] = (15, 15),
        decimation: Tuple[int, int] = (1, 1)) -> Tuple[Optional[Tuple[int, int]], Optional[float]]:
    """
    Perform a single step of the reference search by finding the best matching
    location at given size and scale.

    Parameters
    ----------
    reference_data : BaseReader|numpy.ndarray
    reference_index : None|int
    reference_size : Tuple[int, int]
    moving_data : BaseReader|numpy.ndarray
    moving_index : None|int
    moving_size : Tuple[int, int]
    reference_location : Tuple[int, int]
    moving_location : Tuple[int, int]
    match_box_size : Tuple[int, int]
    moving_deviation : Tuple[int, int]
    decimation : Tuple[int, int]

    Returns
    -------
    best_location : None|Tuple[int, int]
        Will return `None` if there is no information, i.e. the reference patch
        or moving patch is all 0.
    maximum_correlation : None|float
    """

    # NB: we require odd entries here
    match_box_half = (int((match_box_size[0] - 1)/2), int((match_box_size[1] - 1)/2))
    deviation_half = (int((moving_deviation[0] - 1)/2), int((moving_deviation[1] - 1)/2))
    moving_half = (deviation_half[0] + match_box_half[0], deviation_half[1] + match_box_half[1])

    # vet the reference patch location
    ref_box_start = (
        reference_location[0] - match_box_half[0]*decimation[0],
        reference_location[1] - match_box_half[1]*decimation[1])
    if ref_box_start[0] < 0 or \
            ref_box_start[1] < 0 or \
            ref_box_start[0] > reference_size[0] - match_box_size[0]*decimation[0] + 1 or \
            ref_box_start[1] > reference_size[1] - match_box_size[1]*decimation[1] + 1:
        raise ValueError(
            'Overflows bounds. Cannot proceed with reference image of size {}\n\t'
            'using reference box of size {} and decimation {}\n\t'
            'at reference location {}'.format(reference_size, match_box_size, decimation, reference_location))
    ref_box_end = (
        ref_box_start[0] + match_box_size[0]*decimation[0] + 1,
        ref_box_start[1] + match_box_size[1]*decimation[1] + 1)

    # fetch the reference image patch
    if isinstance(reference_data, BaseReader):
        reference_array = reference_data[
                          ref_box_start[0]:ref_box_end[0]:decimation[0],
                          ref_box_start[1]:ref_box_end[1]:decimation[1],
                          reference_index]
    else:
        reference_array = reference_data[
                          ref_box_start[0]:ref_box_end[0]:decimation[0],
                          ref_box_start[1]:ref_box_end[1]:decimation[1]]

    # vet the moving patch location
    mov_loc_temp = [moving_location[0], moving_location[1]]
    for i in [0, 1]:
        if mov_loc_temp[i] < moving_half[i]*decimation[i]:
            mov_loc_temp[i] = moving_half[i]*decimation[i]
        elif mov_loc_temp[i] > moving_size[i] - moving_half[i]*decimation[i] + 1:
            mov_loc_temp[i] = moving_size[i] - moving_half[i]*decimation[i] + 1
    moving_box_start = (
        mov_loc_temp[0] - moving_half[0]*decimation[0],
        mov_loc_temp[1] - moving_half[1]*decimation[1])
    moving_box_end = (
        mov_loc_temp[0] + moving_half[0]*decimation[0] + 1,
        mov_loc_temp[1] + moving_half[1]*decimation[1] + 1)

    # fetch the moving patch array for max correlation check
    if isinstance(moving_data, BaseReader):
        moving_array = moving_data[
                       moving_box_start[0]:moving_box_end[0]:decimation[0],
                       moving_box_start[1]:moving_box_end[1]:decimation[1],
                       moving_index]
    else:
        moving_array = moving_data[
                       moving_box_start[0]:moving_box_end[0]:decimation[0],
                       moving_box_start[1]:moving_box_end[1]:decimation[1]]

    # find the point of maximum correlation
    do_subpixel = numpy.all(decimation == 1)
    best_temp_location, maximum_correlation = _max_correlation_step(reference_array, moving_array, do_subpixel=do_subpixel)

    if best_temp_location is None:
        return best_temp_location, maximum_correlation

    return ((best_temp_location[0])*decimation[0] + mov_loc_temp[0], (best_temp_location[1])*decimation[1] + mov_loc_temp[1]), maximum_correlation


def _single_step_grid(
        reference_data: Union[BaseReader, numpy.ndarray],
        reference_index: Optional[int],
        reference_size: Tuple[int, int],
        moving_data: Union[BaseReader, numpy.ndarray],
        moving_index: Optional[int],
        moving_size: Tuple[int, int],
        reference_box_rough: Tuple[int, int],
        moving_box_rough: Tuple[int, int],
        match_box_size: Tuple[int, int] = (25, 25),
        moving_deviation: Tuple[int, int] = (15, 15),
        decimation: Tuple[int, int] = (1, 1),
        previous_values: Optional[List[List[Dict]]] = None):
    """
    We will determine a series of best matching (small size) patch locations
    between the pixel area of `reference_data` laid out in `reference_box_rough`
    and the pixel area of `moving_data` laid out in `moving_box_rough` - which
    should be very close to the same size.

    Parameters
    ----------
    reference_data : BaseReader|numpy.ndarray
    reference_index : None|int
    reference_size : Tuple[int, int]
    moving_data : BaseReader|numpy.ndarray
    moving_index : None|int
    moving_size : Tuple[int, int]
    reference_box_rough : Tuple[int, int, int, int]
    moving_box_rough : Tuple[int, int, int, int]
    match_box_size : Tuple[int, int]
    moving_deviation : Tuple[int, int]
    decimation : Tuple[int, int]
    previous_values : None|List[List[dict]]

    Returns
    -------
    result_values : List[List[dict]]
        entry `[i][j]` tell the mapping of the reference location in nominal
        reference grid to moving grid location
        :code:`{'reference_location': (row, column),
        'moving_location': (matched_row, matched_column),
        'max_correlation': <value>}`
    """

    def determine_best_guess(ref_point):
        if row_interp is None:
            return (
                ref_point[0] - reference_box_rough[0] + moving_box_rough[0],
                ref_point[1] - reference_box_rough[2] + moving_box_rough[2])
        else:
            return (
                row_interp(ref_point[0]),
                col_interp(ref_point[1]))

    effective_ref_size = (
        int(reference_box_rough[1] - reference_box_rough[0]),
        int(reference_box_rough[3] - reference_box_rough[2]))
    effective_move_size = (
        int(moving_box_rough[1] - moving_box_rough[0]),
        int(moving_box_rough[3] - moving_box_rough[2]))

    # validate the parameters at this scale
    _validate_match_parameters(
        effective_ref_size, effective_move_size, match_box_size, moving_deviation, decimation)

    # construct the grid which we are going to try to map
    half_row_size = int(0.5*(match_box_size[0] - 1)*decimation[0])
    half_col_size = int(0.5*(match_box_size[1] - 1)*decimation[1])
    row_grid = numpy.arange(reference_box_rough[0] + half_row_size, reference_box_rough[1] + half_row_size, 2*half_row_size)
    col_grid = numpy.arange(reference_box_rough[2] + half_col_size, reference_box_rough[3] + half_col_size, 2*half_col_size)

    # create a mapping which estimates (ref_row, ref_col) -> (mov_row, mov_col)
    row_interp = None
    col_interp = None
    if previous_values is not None:
        # determine the approximate derivative values
        _populate_difference_structure(previous_values)

        # get values from our grid, excluding places which require a negative derivative
        ref_locs = []
        mov_locs = []

        for row_values in previous_values:
            for entry in row_values:
                row_der = entry.get('row_derivative', None)
                col_der = entry.get('column_derivative', None)
                if row_der is not None and 0.25 < row_der < 2.0 and \
                        col_der is not None and 0.25 < col_der < 2.0:
                    ref_locs.append(entry['reference_location'])
                    mov_locs.append(entry['moving_location'])

        if len(ref_locs) > 3:
            ref_locs = numpy.array(ref_locs, dtype='float64')
            mov_locs = numpy.array(mov_locs, dtype='float64')
            # create an interpolation function mapping reference coords -> moving coords (so far)
            row_interp = LinearNDInterpolator(ref_locs, mov_locs[:, 0])
            col_interp = LinearNDInterpolator(ref_locs, mov_locs[:, 1])

    result_values = []
    # generate our grid of locations to compare, and then compare
    for row_grid_entry in row_grid:
        the_row_list = []
        result_values.append(the_row_list)

        for col_grid_entry in col_grid:
            ref_loc = (row_grid_entry, col_grid_entry)
            best_loc, max_correlation = _single_step_location(
                reference_data, reference_index, reference_size,
                moving_data, moving_index, moving_size,
                ref_loc, determine_best_guess(ref_loc),
                match_box_size=match_box_size, moving_deviation=moving_deviation, decimation=decimation)
            the_row_list.append(
                {
                    'reference_location': ref_loc,
                    'moving_location': best_loc,
                    'max_correlation': max_correlation
                })

    return result_values


def register_arrays(
        reference_data: numpy.ndarray,
        moving_data: numpy.ndarray) -> List[List[Dict]]:
    """
    Register the moving_data array to the reference_data array using the regi algorithm.

    Parameters
    ----------
    reference_data : numpy.ndarray
    moving_data : numpy.ndarray

    Returns
    -------
    result_values : List[List[dict]]
        entry `[i][j]` tell the mapping of the reference location in nominal
        reference grid to moving grid location
        :code:`{'reference_location': (row, column),
        'moving_location': (matched_row, matched_column),
        'max_correlation': <value>}`
    """

    if not isinstance(reference_data, numpy.ndarray):
        raise TypeError('reference_data must be a numpy array')
    if not isinstance(moving_data, numpy.ndarray):
        raise TypeError('moving_data must be a numpy array')
    if reference_data.ndim != 2:
        raise ValueError('data arrays must be two-dimensional')
    if reference_data.shape != moving_data.shape:
        raise ValueError('data arrays must have the same (2-d) shape')

    the_size = reference_data.shape
    box_rough = (0, the_size[0], 0, the_size[1])
    size_min = min(*the_size)
    if size_min > 2000:
        decimation = (8, 8)
    elif size_min > 1000:
        decimation = (4, 4)
    elif size_min > 500:
        decimation = (2, 2)
    else:
        decimation = (1, 1)

    match_box = (min(25, the_size[0]), min(25, the_size[1]))
    deviation = (min(15, the_size[0]), min(15, the_size[1]))

    return _single_step_grid(
        reference_data, None, the_size,
        moving_data, None, the_size,
        box_rough, box_rough,
        match_box_size=match_box, moving_deviation=deviation, decimation=decimation)



"""
This package mostly centered on base implementations for reader architecture.
"""

__classification__ = 'UNCLASSIFIED'


def open(*args, **kwargs):
    from .converter import open_general
    return open_general(*args, **kwargs)



"""
This module contains the base objects for use in base xml/serializable functionality.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
from xml.etree import ElementTree
import json
from datetime import date, datetime
from collections import OrderedDict
import copy
import re
from io import StringIO
from typing import Dict, Optional

import numpy

from sarpy.compliance import bytes_to_string

try:
    from lxml import etree
except ImportError:
    etree = None


logger = logging.getLogger(__name__)
valid_logger = logging.getLogger('validation')

DEFAULT_STRICT = False


#################
# dom helper functions


def get_node_value(nod: ElementTree.Element) -> Optional[str]:
    """
    XML parsing helper for extracting text value from an ElementTree Element. No error checking performed.

    Parameters
    ----------
    nod : ElementTree.Element
        the xml dom element

    Returns
    -------
    str
        the string value of the node.
    """

    if nod.text is None:
        return None

    val = nod.text.strip()
    if len(val) == 0:
        return None
    else:
        return val


def create_new_node(
        doc: ElementTree.ElementTree,
        tag: str,
        parent: Optional[ElementTree.Element] = None) -> ElementTree.Element:
    """
    XML ElementTree node creation helper function.

    Parameters
    ----------
    doc : ElementTree.ElementTree
        The xml Document object.
    tag : str
        Name/tag for new xml element.
    parent : None|ElementTree.Element
        The parent element for the new element. Defaults to the document root element if unspecified.
    Returns
    -------
    ElementTree.Element
        The new element populated as a child of `parent`.
    """

    if parent is None:
        parent = doc.getroot()  # what if there is no root?
    if parent is None:
        element = ElementTree.Element(tag)
        # noinspection PyProtectedMember, PyUnresolvedReferences
        doc._setroot(element)
        return element
    else:
        return ElementTree.SubElement(parent, tag)


def create_text_node(
        doc: ElementTree.ElementTree,
        tag: str,
        value: str,
        parent: Optional[ElementTree.Element] = None) -> ElementTree.Element:
    """
    XML ElementTree text node creation helper function

    Parameters
    ----------
    doc : ElementTree.ElementTree
        The xml Document object.
    tag : str
        Name/tag for new xml element.
    value : str
        The value for the new element.
    parent : None|ElementTree.Element
        The parent element for the new element. Defaults to the document root element if unspecified.

    Returns
    -------
    ElementTree.Element
        The new element populated as a child of `parent`.
    """

    node = create_new_node(doc, tag, parent=parent)
    node.text = value
    return node


def find_first_child(
        node: ElementTree.Element,
        tag: str,
        xml_ns: Optional[Dict[str, str]],
        ns_key: Optional[str]) -> ElementTree.Element:
    """
    Finds the first child node

    Parameters
    ----------
    node : ElementTree.Element
    tag : str
    xml_ns : None|dict
    ns_key : None|str

    Returns
    -------
    ElementTree.Element
    """

    if xml_ns is None:
        return node.find(tag)
    elif ns_key is None:
        return node.find('default:{}'.format(tag), xml_ns)
    else:
        return node.find('{}:{}'.format(ns_key, tag), xml_ns)


def find_children(node, tag, xml_ns, ns_key):
    """
    Finds the collection of children nodes

    Parameters
    ----------
    node : ElementTree.Element
    tag : str
    xml_ns : None|dict
    ns_key : None|str
    """

    if xml_ns is None:
        return node.findall(tag)
    elif ns_key is None:
        return node.findall('default:{}'.format(tag), xml_ns)
    else:
        return node.findall('{}:{}'.format(ns_key, tag), xml_ns)


def parse_xml_from_string(xml_string):
    """
    Parse the ElementTree root node and xml namespace dict from a xml string.

    Parameters
    ----------
    xml_string : str|bytes

    Returns
    -------
    root_node: ElementTree.Element
    xml_ns: Dict[str, str]
    """

    xml_string = bytes_to_string(xml_string, encoding='utf-8')

    root_node = ElementTree.fromstring(xml_string)
    # define the namespace dictionary
    xml_ns = dict([node for _, node in ElementTree.iterparse(StringIO(xml_string), events=('start-ns',))])
    if len(xml_ns.keys()) == 0:
        xml_ns = None
    elif '' in xml_ns:
        xml_ns['default'] = xml_ns['']
    else:
        # default will be the namespace for the root node
        namespace_match = re.match(r'\{.*\}', root_node.tag)
        if namespace_match is None:
            raise ValueError('Trouble finding the default namespace for tag {}'.format(root_node.tag))
        xml_ns['default'] = namespace_match[0][1:-1]
    return root_node, xml_ns


def parse_xml_from_file(xml_file_path):
    """
    Parse the ElementTree root node and xml namespace dict from a xml file.

    Parameters
    ----------
    xml_file_path : str

    Returns
    -------
    root_node: ElementTree.Element
    xml_ns: Dict[str, str]
    """

    with open(xml_file_path, 'rb') as fi:
        xml_bytes = fi.read()
    return parse_xml_from_string(xml_bytes)


def validate_xml_from_string(xml_string, xsd_path, output_logger=None):
    """
    Validate a xml string against a given xsd document.

    Parameters
    ----------
    xml_string : str|bytes
    xsd_path : str
        The path to the relevant xsd document.
    output_logger
        A desired output logger.

    Returns
    -------
    bool
        `True` if valid, `False` otherwise. Failure reasons will be
        logged at `'error'` level by the module.
    """

    if etree is None:
        raise ImportError(
            'The lxml package was not successfully imported,\n\t'
            'and this xml validation requires lxml.')

    xml_doc = etree.fromstring(xml_string)
    xml_schema = etree.XMLSchema(file=xsd_path)
    validity = xml_schema.validate(xml_doc)
    if not validity:
        for entry in xml_schema.error_log:
            msg = 'XML validation error on line {}\n\t{}'.format(
                entry.line, entry.message.encode('utf-8'))
            if output_logger is None:
                logger.error(msg)
            else:
                output_logger.error(msg)
    return validity


def validate_xml_from_file(xml_path, xsd_path, output_logger=None):
    """
    Validate a xml string against a given xsd document.

    Parameters
    ----------
    xml_path : str
        The path to the relevant xml file
    xsd_path : str
        The path to the relevant xsd document.
    output_logger
        A desired output logger.

    Returns
    -------
    bool
        `True` if valid, `False` otherwise. Failure reasons will be
        logged at `'error'` level by the module.
    """

    with open(xml_path, 'rb') as fi:
        xml_bytes = fi.read()

    return validate_xml_from_string(xml_bytes, xsd_path, output_logger=output_logger)


###
# parsing functions - for reusable functionality in descriptors or other property definitions


def parse_str(value, name, instance):
    if value is None:
        return None
    if isinstance(value, str):
        return value
    elif isinstance(value, ElementTree.Element):
        node_value = get_node_value(value)
        return "" if node_value is None else node_value
    else:
        raise TypeError(
            'field {} of class {} requires a string value.'.format(name, instance.__class__.__name__))


def parse_bool(value, name, instance):
    def parse_string(val):
        if val.lower() in ['0', 'false']:
            return False
        elif val.lower() in ['1', 'true']:
            return True
        else:
            raise ValueError(
                'Boolean field {} of class {} cannot assign from string value {}. '
                'It must be one of ["0", "false", "1", "true"]'.format(name, instance.__class__.__name__, val))

    if value is None:
        return None
    if isinstance(value, bool):
        return value
    elif isinstance(value, int) or isinstance(value, numpy.bool_):
        return bool(value)
    elif isinstance(value, ElementTree.Element):
        # from XML deserialization
        return parse_string(get_node_value(value))
    elif isinstance(value, str):
        return parse_string(value)
    else:
        raise ValueError('Boolean field {} of class {} cannot assign from type {}.'.format(
            name, instance.__class__.__name__, type(value)))


def parse_int(value, name, instance):
    if value is None:
        return None
    if isinstance(value, int):
        return value
    elif isinstance(value, ElementTree.Element):
        # from XML deserialization
        return parse_int(get_node_value(value), name, instance)
    elif isinstance(value, str):
        try:
            return int(value)
        except ValueError as e:
            logger.warning(
                'Got non-integer value {}\n\t'
                'for integer valued field {} of class {}'.format(
                    value, name, instance.__class__.__name__))
            # noinspection PyBroadException
            try:
                return int(float(value))
            except Exception:
                raise e
    else:
        # user or json deserialization
        return int(value)


# noinspection PyUnusedLocal
def parse_float(value, name, instance):
    if value is None:
        return None
    if isinstance(value, float):
        return value
    elif isinstance(value, ElementTree.Element):
        # from XML deserialization
        return float(get_node_value(value))
    else:
        # user or json deserialization
        return float(value)


def parse_complex(value, name, instance):
    if value is None:
        return None
    if isinstance(value, complex):
        return value
    elif isinstance(value, ElementTree.Element):
        xml_ns = getattr(instance, '_xml_ns', None)
        # noinspection PyProtectedMember
        if hasattr(instance, '_child_xml_ns_key') and name in instance._child_xml_ns_key:
            # noinspection PyProtectedMember
            xml_ns_key = instance._child_xml_ns_key[name]
        else:
            xml_ns_key = getattr(instance, '_xml_ns_key', None)
        # from XML deserialization
        rnode = find_children(value, 'Real', xml_ns, xml_ns_key)
        inode = find_children(value, 'Imag', xml_ns, xml_ns_key)

        if len(rnode) != 1:
            raise ValueError(
                'There must be exactly one Real component of a complex type node '
                'defined for field {} of class {}.'.format(name, instance.__class__.__name__))
        if len(inode) != 1:
            raise ValueError(
                'There must be exactly one Imag component of a complex type node '
                'defined for field {} of class {}.'.format(name, instance.__class__.__name__))
        real = float(get_node_value(rnode[0]))
        imag = float(get_node_value(inode[0]))
        return complex(real, imag)
    elif isinstance(value, dict):
        # from json deserialization
        real = None
        for key in ['re', 'real', 'Real']:
            real = value.get(key, real)
        imag = None
        for key in ['im', 'imag', 'Imag']:
            imag = value.get(key, imag)
        if real is None or imag is None:
            raise ValueError(
                'Cannot convert dict {} to a complex number for field {} of '
                'class {}.'.format(value, name, instance.__class__.__name__))
        return complex(real, imag)
    else:
        # from user - I can't imagine that this would ever work
        return complex(value)


def parse_datetime(value, name, instance, units='us'):
    if value is None:
        return None
    if isinstance(value, numpy.datetime64):
        return value
    elif isinstance(value, str):
        # handle Z timezone identifier explicitly - any timezone identifier is deprecated
        if value[-1] == 'Z':
            return numpy.datetime64(value[:-1], units)
        else:
            return numpy.datetime64(value, units)
    elif isinstance(value, ElementTree.Element):
        # from XML deserialization - extract the string
        return parse_datetime(get_node_value(value), name, instance, units=units)
    elif isinstance(value, (date, datetime, numpy.int64, numpy.float64)):
        return numpy.datetime64(value, units)
    elif isinstance(value, int):
        # this is less safe, because the units are unknown...
        return numpy.datetime64(value, units)
    else:
        raise TypeError(
            'Field {} for class {} expects datetime convertible input, and '
            'got {}'.format(name, instance.__class__.__name__, type(value)))


def parse_serializable(value, name, instance, the_type):
    if value is None:
        return None
    if isinstance(value, the_type):
        return value
    elif isinstance(value, dict):
        return the_type.from_dict(value)
    elif isinstance(value, ElementTree.Element):
        xml_ns = getattr(instance, '_xml_ns', None)
        if hasattr(instance, '_child_xml_ns_key'):
            # noinspection PyProtectedMember
            xml_ns_key = instance._child_xml_ns_key.get(name, getattr(instance, '_xml_ns_key', None))
        else:
            xml_ns_key = getattr(instance, '_xml_ns_key', None)
        return the_type.from_node(value, xml_ns, ns_key=xml_ns_key)
    elif isinstance(value, (numpy.ndarray, list, tuple)):
        if issubclass(the_type, Arrayable):
            return the_type.from_array(value)
        else:
            raise TypeError(
                'Field {} of class {} is of type {} (not a subclass of Arrayable) and '
                'got an argument of type {}.'.format(name, instance.__class__.__name__, the_type, type(value)))
    else:
        raise TypeError(
            'Field {} of class {} is expecting type {}, but got an instance of incompatible '
            'type {}.'.format(name, instance.__class__.__name__, the_type, type(value)))


def parse_serializable_array(value, name, instance, child_type, child_tag):
    if value is None:
        return None
    if isinstance(value, child_type):
        # this is the child element
        return numpy.array([value, ], dtype='object')
    elif isinstance(value, numpy.ndarray):
        if value.dtype.name != 'object':
            if issubclass(child_type, Arrayable):
                return numpy.array([child_type.from_array(array) for array in value], dtype='object')
            else:
                raise ValueError(
                    'Attribute {} of array type functionality belonging to class {} got an ndarray of dtype {},'
                    'and child type is not a subclass of Arrayable.'.format(
                        name, instance.__class__.__name__, value.dtype))
        elif len(value.shape) != 1:
            raise ValueError(
                'Attribute {} of array type functionality belonging to class {} got an ndarray of shape {},'
                'but requires a one dimensional array.'.format(
                    name, instance.__class__.__name__, value.shape))
        elif not isinstance(value[0], child_type):
            raise TypeError(
                'Attribute {} of array type functionality belonging to class {} got an ndarray containing '
                'first element of incompatible type {}.'.format(
                    name, instance.__class__.__name__, type(value[0])))
        return value
    elif isinstance(value, ElementTree.Element):
        xml_ns = getattr(instance, '_xml_ns', None)
        if hasattr(instance, '_child_xml_ns_key'):
            # noinspection PyProtectedMember
            xml_ns_key = instance._child_xml_ns_key.get(name, getattr(instance, '_xml_ns_key', None))
        else:
            xml_ns_key = getattr(instance, '_xml_ns_key', None)
        # this is the parent node from XML deserialization
        size = int(value.attrib.get('size', -1))  # NB: Corner Point arrays don't have
        # extract child nodes at top level
        child_nodes = find_children(value, child_tag, xml_ns, xml_ns_key)

        if size == -1:  # fill in, if it's missing
            size = len(child_nodes)
        if len(child_nodes) != size:
            raise ValueError(
                'Attribute {} of array type functionality belonging to class {} got a ElementTree element '
                'with size attribute {}, but has {} child nodes with tag {}.'.format(
                    name, instance.__class__.__name__, size, len(child_nodes), child_tag))
        new_value = numpy.empty((size, ), dtype='object')
        for i, entry in enumerate(child_nodes):
            new_value[i] = child_type.from_node(entry, xml_ns, ns_key=xml_ns_key)
        return new_value
    elif isinstance(value, (list, tuple)):
        # this would arrive from users or json deserialization
        if len(value) == 0:
            return numpy.empty((0,), dtype='object')
        elif isinstance(value[0], child_type):
            return numpy.array(value, dtype='object')
        elif isinstance(value[0], dict):
            # NB: charming errors are possible here if something stupid has been done.
            return numpy.array([child_type.from_dict(node) for node in value], dtype='object')
        elif isinstance(value[0], (numpy.ndarray, list, tuple)):
            if issubclass(child_type, Arrayable):
                return numpy.array([child_type.from_array(array) for array in value], dtype='object')
            elif hasattr(child_type, 'Coefs'):
                return numpy.array([child_type(Coefs=array) for array in value], dtype='object')
            else:
                raise ValueError(
                    'Attribute {} of array type functionality belonging to class {} got an list '
                    'containing elements type {} and construction failed.'.format(
                        name, instance.__class__.__name__, type(value[0])))
        else:
            raise TypeError(
                'Attribute {} of array type functionality belonging to class {} got a list containing first '
                'element of incompatible type {}.'.format(name, instance.__class__.__name__, type(value[0])))
    else:
        raise TypeError(
            'Attribute {} of array type functionality belonging to class {} got incompatible type {}.'.format(
                name, instance.__class__.__name__, type(value)))


def parse_serializable_list(value, name, instance, child_type):
    if value is None:
        return None
    if isinstance(value, child_type):
        # this is the child element
        return [value, ]

    xml_ns = getattr(instance, '_xml_ns', None)
    if hasattr(instance, '_child_xml_ns_key'):
        # noinspection PyProtectedMember
        xml_ns_key = instance._child_xml_ns_key.get(name, getattr(instance, '_xml_ns_key', None))
    else:
        xml_ns_key = getattr(instance, '_xml_ns_key', None)
    if isinstance(value, ElementTree.Element):
        # this is the child
        return [child_type.from_node(value, xml_ns, ns_key=xml_ns_key), ]
    elif isinstance(value, list) or isinstance(value[0], child_type):
        if len(value) == 0:
            return value
        elif isinstance(value[0], child_type):
            return [entry for entry in value]
        elif isinstance(value[0], dict):
            # NB: charming errors are possible if something stupid has been done.
            return [child_type.from_dict(node) for node in value]
        elif isinstance(value[0], ElementTree.Element):
            return [child_type.from_node(node, xml_ns, ns_key=xml_ns_key) for node in value]
        else:
            raise TypeError(
                'Field {} of list type functionality belonging to class {} got a '
                'list containing first element of incompatible type '
                '{}.'.format(name, instance.__class__.__name__, type(value[0])))
    else:
        raise TypeError(
            'Field {} of class {} got incompatible type {}.'.format(
                name, instance.__class__.__name__, type(value)))


def parse_parameters_collection(value, name, instance):
    if value is None:
        return None
    if isinstance(value, dict):
        return value
    elif isinstance(value, list):
        out = OrderedDict()
        if len(value) == 0:
            return out
        if isinstance(value[0], ElementTree.Element):
            for entry in value:
                out[entry.attrib['name']] = get_node_value(entry)
            return out
        else:
            raise TypeError(
                'Field {} of list type functionality belonging to class {} got a '
                'list containing first element of incompatible type '
                '{}.'.format(name, instance.__class__.__name__, type(value[0])))
    else:
        raise TypeError(
            'Field {} of class {} got incompatible type {}.'.format(
                name, instance.__class__.__name__, type(value)))


##################
# Main class defining structure

class Serializable(object):
    """
    Basic abstract class specifying the serialization pattern. There are no clearly defined Python conventions
    for this issue. Every effort has been made to select sensible choices, but this is an individual effort.

    Notes
    -----
        All fields MUST BE LISTED in the `_fields` tuple. Everything listed in `_required` tuple will be checked
        for inclusion in `_fields` tuple. Note that special care must be taken to ensure compatibility of `_fields`
        tuple, if inheriting from an extension of this class.
    """
    _fields = ()
    """collection of field names"""
    _required = ()
    """subset of `_fields` defining the required (for the given object, according to the sicd standard) fields"""
    _tag_override = {}
    """On occasion, the xml tag and the corresponding variable name may need to differ. 
    This dictionary should be populated as `{<variable name> : <tag name>}`."""

    _collections_tags = {}
    """
    Entries only appropriate for list/array type objects. Entry formatting:

    * `{'array': True, 'child_tag': <child_name>}` represents an array object, which will have int attribute `size`.
      It has *size* children with tag=<child_name>, each of which has an attribute `index`, which is not always an
      integer. Even when it is an integer, it apparently sometimes follows the matlab convention (1 based), and
      sometimes the standard convention (0 based). In this case, I will deserialize as though the objects are
      properly ordered, and the deserialized objects will have the `index` property from the xml, but it will not
      be used to determine array order - which will be simply carried over from file order.

    * `{'array': False, 'child_tag': <child_name>}` represents a collection of things with tag=<child_name>.
      This entries are not directly below one coherent container tag, but just dumped into an object.
      For example of such usage search for "Parameter" in the SICD standard.

      In this case, I've have to create an ephemeral variable in the class that doesn't exist in the standard,
      and it's not clear what the intent is for this unstructured collection, so used a list object.
      For example, I have a variable called `Parameters` in `CollectionInfoType`, whose job it to contain the
      parameters objects.
    """

    _numeric_format = {}
    """define dict entries of numeric formatting for serialization"""
    _set_as_attribute = ()
    """serialize these fields as xml attributes"""
    _choice = ()
    """
    Entries appropriate for choice selection between attributes. Entry formatting:

    * `{'required': True, 'collection': <tuple of attribute names>}` - indicates that EXACTLY only one of the 
      attributes should be populated.

    * `{'required': False, 'collection': <tuple of attribute names>}` - indicates that no more than one of the 
      attributes should be populated.
    """
    _child_xml_ns_key = {}
    """
    The expected namespace key for attributes. No entry indicates the default namespace. 
    This is important for SIDD handling, but not required for SICD handling.
    """

    # NB: it may be good practice to use __slots__ to further control class functionality?

    def __init__(self, **kwargs):
        """
        The default constructor. For each attribute name in `self._fields`, fetches the value (or None) from
        the `kwargs` dict, and sets the class instance attribute value. The details for attribute value validation,
        present for virtually every attribute, will be implemented specifically as descriptors.

        Parameters
        ----------
        **kwargs :
            the keyword arguments dictionary - the possible entries match the attributes.
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        unexpected_args = [key for key in kwargs if key not in self._fields and key[0] != '_']
        if len(unexpected_args) > 0:
            raise ValueError(
                'Received unexpected construction argument {} for attribute '
                'collection {}'.format(unexpected_args, self._fields))

        for attribute in self._fields:
            if attribute in kwargs:
                try:
                    setattr(self, attribute, kwargs.get(attribute, None))
                except AttributeError:
                    # NB: this is included to allow for read only properties without breaking the paradigm
                    #   Silently catching errors can potentially cover up REAL issues.
                    pass

    def __str__(self):
        return '{}(**{})'.format(self.__class__.__name__, json.dumps(self.to_dict(check_validity=False), indent=1))

    def __repr__(self):
        return '{}(**{})'.format(self.__class__.__name__, self.to_dict(check_validity=False))

    def __setattr__(self, key, value):
        if not (key.startswith('_') or (key in self._fields) or hasattr(self.__class__, key) or hasattr(self, key)):
            # not expected attribute - descriptors, properties, etc
            logger.warning(
                'Class {} instance receiving unexpected attribute {}.\n\t'
                'Ensure that this is not a typo of an expected field name.'.format(self.__class__.__name__, key))
        object.__setattr__(self, key, value)

    def __getstate__(self):
        """
        Method for allowing copying and/or pickling of state.

        Returns
        -------
        dict
            The dict representation for the object.
        """

        return self.to_dict(check_validity=False, strict=False)

    def __setstate__(self, the_dict):
        """
        Method for reconstructing from the serialized state.
        """

        return self.__init__(**the_dict)

    def set_numeric_format(self, attribute, format_string):
        """Sets the numeric format string for the given attribute.

        Parameters
        ----------
        attribute : str
            attribute for which the format applies - must be in `_fields`.
        format_string : str
            format string to be applied
        Returns
        -------
        None
        """
        # Extend this to include format function capabilities. Maybe numeric_format is not the right name?
        if attribute not in self._fields:
            raise ValueError('attribute {} is not permitted for class {}'.format(attribute, self.__class__.__name__))
        self._numeric_format[attribute] = format_string

    def _get_formatter(self, attribute):
        """Return a formatting function for the given attribute. This will default to `str` if no other
        option is presented.

        Parameters
        ----------
        attribute : str
            the given attribute name

        Returns
        -------
        Callable
            format function
        """

        entry = self._numeric_format.get(attribute, None)
        if isinstance(entry, str):
            fmt_str = '{0:' + entry + '}'
            return fmt_str.format
        elif callable(entry):
            return entry
        else:
            return str

    def log_validity_error(self, msg):
        """
        Log a validity check error message.

        Parameters
        ----------
        msg : str
        """

        valid_logger.error('{}: {}'.format(self.__class__.__name__, msg))

    def log_validity_warning(self, msg):
        """
        Log a validity check warning message.

        Parameters
        ----------
        msg : str
        """

        valid_logger.warning('{}: {}'.format(self.__class__.__name__, msg))

    def log_validity_info(self, msg):
        """
        Log a validation info message.

        Parameters
        ----------
        msg : str
        """

        valid_logger.info('{}: {}'.format(self.__class__.__name__, msg))

    def is_valid(self, recursive=False, stack=False):
        """Returns the validity of this object according to the schema. This is done by inspecting that all required
        fields (i.e. entries of `_required`) are not `None`.

        Parameters
        ----------
        recursive : bool
            True if we recursively check that child are also valid. This may result in verbose (i.e. noisy) logging.
        stack : bool
            Print a recursive error message?

        Returns
        -------
        bool
            condition for validity of this element
        """

        all_required = self._basic_validity_check()
        if not recursive:
            return all_required

        valid_children = self._recursive_validity_check(stack=stack)
        return all_required & valid_children

    def _basic_validity_check(self):
        """
        Perform the basic validity check on the direct attributes with no recursive checking.

        Returns
        -------
        bool
             True if all requirements *AT THIS LEVEL* are satisfied, otherwise False.
        """

        all_required = True
        for attribute in self._required:
            present = (getattr(self, attribute) is not None)
            if not present:
                self.log_validity_error("Missing required attribute {}".format(attribute))
            all_required &= present

        choices = True
        for entry in self._choice:
            required = entry.get('required', False)
            collect = entry['collection']
            # verify that no more than one of the entries in collect is set.
            present = []
            for attribute in collect:
                if getattr(self, attribute) is not None:
                    present.append(attribute)
            if len(present) == 0 and required:
                self.log_validity_error(
                    "Exactly one of the attributes {} should be set, but none are set".format(collect))
                choices = False
            elif len(present) > 1:
                self.log_validity_error(
                    "Exactly one of the attributes {} should be set, but multiple ({}) are set".format(collect,
                                                                                                       present))
                choices = False

        return all_required and choices

    def _recursive_validity_check(self, stack=False):
        """
        Perform a recursive validity check on all present attributes.

        Parameters
        ----------
        stack : bool
            Print a recursive error message?

        Returns
        -------
        bool
             True if requirements are recursively satisfied *BELOW THIS LEVEL*, otherwise False.
        """

        def check_item(value):
            if isinstance(value, (Serializable, SerializableArray)):
                return value.is_valid(recursive=True, stack=stack)
            return True

        valid_children = True
        for attribute in self._fields:
            val = getattr(self, attribute)
            good = True
            if isinstance(val, (Serializable, SerializableArray)):
                good = check_item(val)
            elif isinstance(val, list):
                for entry in val:
                    good &= check_item(entry)
            # any issues will be logged as discovered, but should we help with the "stack"?
            if not good and stack:
                self.log_validity_error(
                    "Issue discovered with attribute {} of type {}.".format(attribute, type(val)))
            valid_children &= good
        return valid_children

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        """For XML deserialization.

        Parameters
        ----------
        node : ElementTree.Element
            dom element for serialized class instance
        xml_ns : None|dict
            The xml namespace dictionary.
        ns_key : None|str
            The xml namespace key. If `xml_ns` is None, then this is ignored. If `None` and `xml_ns` is not None,
            then the string `default` will be used. This will be recursively passed down,
            unless overridden by an entry of the cls._child_xml_ns_key dictionary.
        kwargs : None|dict
            `None` or dictionary of previously serialized attributes. For use in inheritance call, when certain
            attributes require specific deserialization.
        Returns
        -------
            Corresponding class instance
        """

        if len(node) == 0 and len(node.attrib) == 0:
            logger.warning(
                'There are no children or attributes associated\n\t'
                'with node {}\n\t'
                'for class {}.'.format(node, cls))
            # return None

        def handle_attribute(the_attribute, the_tag, the_xml_ns_key):
            if the_xml_ns_key is not None:  # handle namespace, if necessary
                fetch_tag = '{' + xml_ns[the_xml_ns_key] + '}' + the_tag
            else:
                fetch_tag = the_tag
            kwargs[the_attribute] = node.attrib.get(fetch_tag, None)

        def handle_single(the_attribute, the_tag, the_xml_ns_key):
            kwargs[the_attribute] = find_first_child(node, the_tag, xml_ns, the_xml_ns_key)

        def handle_list(attrib, ch_tag, the_xml_ns_key):
            cnodes = find_children(node, ch_tag, xml_ns, the_xml_ns_key)
            if len(cnodes) > 0:
                kwargs[attrib] = cnodes

        if kwargs is None:
            kwargs = {}
        kwargs['_xml_ns'] = xml_ns
        kwargs['_xml_ns_key'] = ns_key

        if not isinstance(kwargs, dict):
            raise ValueError(
                "Named input argument kwargs for class {} must be dictionary instance".format(cls))

        for attribute in cls._fields:
            if attribute in kwargs:
                continue

            kwargs[attribute] = None
            # This value will be replaced if tags are present
            # Note that we want to try explicitly setting to None to trigger descriptor behavior
            # for required fields (warning or error)

            base_tag_name = cls._tag_override.get(attribute, attribute)

            # determine any expected xml namespace for the given entry
            if attribute in cls._child_xml_ns_key:
                xml_ns_key = cls._child_xml_ns_key[attribute]
            else:
                xml_ns_key = ns_key
            # verify that the xml namespace will work
            if xml_ns_key is not None:
                if xml_ns is None:
                    raise ValueError('Attribute {} in class {} expects a xml namespace entry of {}, '
                                     'but xml_ns is None.'.format(attribute, cls, xml_ns_key))
                elif xml_ns_key not in xml_ns:
                    raise ValueError('Attribute {} in class {} expects a xml namespace entry of {}, '
                                     'but xml_ns does not contain this key.'.format(attribute, cls, xml_ns_key))

            if attribute in cls._set_as_attribute:
                xml_ns_key = cls._child_xml_ns_key.get(attribute, None)
                handle_attribute(attribute, base_tag_name, xml_ns_key)
            elif attribute in cls._collections_tags:
                # it's a collection type parameter
                array_tag = cls._collections_tags[attribute]
                array = array_tag.get('array', False)
                child_tag = array_tag.get('child_tag', None)
                if array:
                    handle_single(attribute, base_tag_name, xml_ns_key)
                elif child_tag is not None:
                    handle_list(attribute, child_tag, xml_ns_key)
                else:
                    # the metadata is broken
                    raise ValueError(
                        'Attribute {} in class {} is listed in the _collections_tags dictionary, but the '
                        '`child_tag` value is either not populated or None.'.format(attribute, cls))
            else:
                # it's a regular property
                handle_single(attribute, base_tag_name, xml_ns_key)
        return cls.from_dict(kwargs)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        """For XML serialization, to a dom element.

        Parameters
        ----------
        doc : ElementTree.ElementTree
            The xml Document
        tag : None|str
            The tag name. Defaults to the value of `self._tag` and then the class name if unspecified.
        ns_key : None|str
            The namespace prefix. This will be recursively passed down, unless overridden by an entry in the
            _child_xml_ns_key dictionary.
        parent : None|ElementTree.Element
            The parent element. Defaults to the document root element if unspecified.
        check_validity : bool
            Check whether the element is valid before serializing, by calling :func:`is_valid`.
        strict : bool
            Only used if `check_validity = True`. In that case, if `True` then raise an
            Exception (of appropriate type) if the structure is not valid, if `False` then log a
            hopefully helpful message.
        exclude : tuple
            Attribute names to exclude from this generic serialization. This allows for child classes
            to provide specific serialization for special properties, after using this super method.

        Returns
        -------
        ElementTree.Element
            The constructed dom element, already assigned to the parent element.
        """

        def serialize_attribute(node, the_tag, val, format_function, the_xml_ns_key):
            if the_xml_ns_key is None or the_xml_ns_key == 'default':
                node.attrib[the_tag] = format_function(val)
            else:
                node.attrib['{}:{}'.format(the_xml_ns_key, the_tag)] = format_function(val)

        def serialize_array(node, the_tag, ch_tag, val, format_function, size_attrib, the_xml_ns_key):
            if not isinstance(val, numpy.ndarray):
                # this should really never happen, unless someone broke the class badly by fiddling with
                # _collections_tag or the descriptor at runtime
                raise TypeError(
                    'The value associated with attribute {} is an instance of class {} should be an array based on '
                    'the metadata in the _collections_tags dictionary, but we received an instance of '
                    'type {}'.format(attribute, self.__class__.__name__, type(val)))
            if not len(val.shape) == 1:
                # again, I have no idea how we'd find ourselves here, unless inconsistencies have been introduced
                # into the descriptor
                raise ValueError(
                    'The value associated with attribute {} is an instance of class {}, if None, is required to be'
                    'a one-dimensional numpy.ndarray, but it has shape {}'.format(
                        attribute, self.__class__.__name__, val.shape))
            if val.size == 0:
                return  # serializing an empty array is dumb

            if val.dtype.name == 'float64':
                if the_xml_ns_key is None:
                    anode = create_new_node(doc, the_tag, parent=node)
                else:
                    anode = create_new_node(doc, '{}:{}'.format(the_xml_ns_key, the_tag), parent=node)
                anode.attrib[size_attrib] = str(val.size)
                for i, val in enumerate(val):
                    vnode = create_text_node(doc, ch_tag, format_function(val), parent=anode)
                    vnode.attrib['index'] = str(i) if ch_tag == 'Amplitude' else str(i + 1)
            else:
                # I have no idea how we'd find ourselves here, unless inconsistencies have been introduced
                # into the descriptor
                raise ValueError(
                    'The value associated with attribute {} is an instance of class {}, if None, is required to be'
                    'a numpy.ndarray of dtype float64 or object, but it has dtype {}'.format(
                        attribute, self.__class__.__name__, val.dtype))

        def serialize_list(node, ch_tag, val, format_function, the_xml_ns_key):
            if not isinstance(val, list):
                # this should really never happen, unless someone broke the class badly by fiddling with
                # _collections_tags or the descriptor?
                raise TypeError(
                    'The value associated with attribute {} is an instance of class {} should be a list based on '
                    'the metadata in the _collections_tags dictionary, but we received an instance of '
                    'type {}'.format(attribute, self.__class__.__name__, type(val)))
            if len(val) == 0:
                return  # serializing an empty list is dumb
            else:
                for entry in val:
                    if entry is not None:
                        serialize_plain(node, ch_tag, entry, format_function, the_xml_ns_key)

        def serialize_plain(node, the_tag, val, format_function, the_xml_ns_key):
            # may be called not at top level - if object array or list is present
            prim_tag = '{}:{}'.format(the_xml_ns_key, the_tag) if the_xml_ns_key is not None else the_tag
            if isinstance(val, (Serializable, SerializableArray)):
                val.to_node(doc, the_tag, ns_key=the_xml_ns_key, parent=node,
                            check_validity=check_validity, strict=strict)
            elif isinstance(val, ParametersCollection):
                val.to_node(doc, ns_key=the_xml_ns_key, parent=node, check_validity=check_validity, strict=strict)
            elif isinstance(val, bool):  # this must come before int, where it would evaluate as true
                create_text_node(doc, prim_tag, 'true' if val else 'false', parent=node)
            elif isinstance(val, str):
                create_text_node(doc, prim_tag, val, parent=node)
            elif isinstance(val, int):
                create_text_node(doc, prim_tag, format_function(val), parent=node)
            elif isinstance(val, float):
                create_text_node(doc, prim_tag, format_function(val), parent=node)
            elif isinstance(val, numpy.datetime64):
                out2 = str(val)
                out2 = out2 + 'Z' if out2[-1] != 'Z' else out2
                create_text_node(doc, prim_tag, out2, parent=node)
            elif isinstance(val, complex):
                cnode = create_new_node(doc, prim_tag, parent=node)
                if the_xml_ns_key is None:
                    create_text_node(doc, 'Real', format_function(val.real), parent=cnode)
                    create_text_node(doc, 'Imag', format_function(val.imag), parent=cnode)
                else:
                    create_text_node(doc, '{}:Real'.format(the_xml_ns_key), format_function(val.real), parent=cnode)
                    create_text_node(doc, '{}:Imag'.format(the_xml_ns_key), format_function(val.imag), parent=cnode)
            elif isinstance(val, date):  # should never exist at present
                create_text_node(doc, prim_tag, val.isoformat(), parent=node)
            elif isinstance(val, datetime):  # should never exist at present
                create_text_node(doc, prim_tag, val.isoformat(sep='T'), parent=node)
            else:
                raise ValueError(
                    'An entry for class {} using tag {} is of type {},\n'
                    'and serialization has not been implemented'.format(self.__class__.__name__, the_tag, type(val)))

        if check_validity:
            if not self.is_valid(stack=False):
                msg = "{} is not valid,\n\t" \
                      "and cannot be SAFELY serialized to XML according to the " \
                      "SICD standard.".format(self.__class__.__name__)
                if strict:
                    raise ValueError(msg)
                logger.warning(msg)
        # create the main node
        if (ns_key is not None and ns_key != 'default') and not tag.startswith(ns_key + ':'):
            nod = create_new_node(doc, '{}:{}'.format(ns_key, tag), parent=parent)
        else:
            nod = create_new_node(doc, tag, parent=parent)

        # serialize the attributes
        for attribute in self._fields:
            if attribute in exclude:
                continue

            value = getattr(self, attribute)
            if value is None:
                continue
            fmt_func = self._get_formatter(attribute)
            base_tag_name = self._tag_override.get(attribute, attribute)
            if attribute in self._set_as_attribute:
                xml_ns_key = self._child_xml_ns_key.get(attribute, ns_key)
                serialize_attribute(nod, base_tag_name, value, fmt_func, xml_ns_key)
            else:
                # should we be using some namespace?
                if attribute in self._child_xml_ns_key:
                    xml_ns_key = self._child_xml_ns_key[attribute]
                else:
                    xml_ns_key = getattr(self, '_xml_ns_key', ns_key)
                    if xml_ns_key == 'default':
                        xml_ns_key = None

                if isinstance(value, (numpy.ndarray, list)):
                    array_tag = self._collections_tags.get(attribute, None)
                    if array_tag is None:
                        raise AttributeError(
                            'The value associated with attribute {} in an instance of class {} is of type {}, '
                            'but nothing is populated in the _collection_tags dictionary.'.format(
                                attribute, self.__class__.__name__, type(value)))
                    child_tag = array_tag.get('child_tag', None)
                    if child_tag is None:
                        raise AttributeError(
                            'The value associated with attribute {} in an instance of class {} is of type {}, '
                            'but `child_tag` is not populated in the _collection_tags dictionary.'.format(
                                attribute, self.__class__.__name__, type(value)))
                    size_attribute = array_tag.get('size_attribute', 'size')
                    if isinstance(value, numpy.ndarray):
                        serialize_array(nod, base_tag_name, child_tag, value, fmt_func, size_attribute, xml_ns_key)
                    else:
                        serialize_list(nod, child_tag, value, fmt_func, xml_ns_key)
                else:
                    serialize_plain(nod, base_tag_name, value, fmt_func, xml_ns_key)
        return nod

    @classmethod
    def from_dict(cls, input_dict):
        """For json deserialization, from dict instance.

        Parameters
        ----------
        input_dict : dict
            Appropriate parameters dict instance for deserialization

        Returns
        -------
            Corresponding class instance
        """

        return cls(**input_dict)

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        """
        For json serialization.

        Parameters
        ----------
        check_validity : bool
            Check whether the element is valid before serializing, by calling :func:`is_valid`.
        strict : bool
            Only used if `check_validity = True`. In that case, if `True` then raise an
            Exception (of appropriate type) if the structure is not valid, if `False` then log a
            hopefully helpful message.
        exclude : tuple
            Attribute names to exclude from this generic serialization. This allows for child classes
            to provide specific serialization for special properties, after using this super method.

        Returns
        -------
        OrderedDict
            dict representation of class instance appropriate for direct json serialization.
        """

        # noinspection PyUnusedLocal
        def serialize_array(ch_tag, val):
            if not len(val.shape) == 1:
                # again, I have no idea how we'd find ourselves here, unless inconsistencies have been introduced
                # into the descriptor
                raise ValueError(
                    'The value associated with attribute {} is an instance of class {}, if None, is required to be'
                    'a one-dimensional numpy.ndarray, but it has shape {}'.format(
                        attribute, self.__class__.__name__, val.shape))

            if val.size == 0:
                return []

            if val.dtype.name == 'float64':
                return [float(el) for el in val]
            else:
                # I have no idea how we'd find ourselves here, unless inconsistencies have been introduced
                # into the descriptor
                raise ValueError(
                    'The value associated with attribute {} is an instance of class {}. This is expected to be'
                    'a numpy.ndarray of dtype float64, but it has dtype {}'.format(
                        attribute, self.__class__.__name__, val.dtype))

        def serialize_list(ch_tag, val):
            if len(val) == 0:
                return []
            else:
                return [serialize_plain(ch_tag, entry) for entry in val if entry is not None]

        def serialize_plain(field, val):
            # may be called not at top level - if object array or list is present
            if isinstance(val, Serializable):
                return val.to_dict(check_validity=check_validity, strict=strict)
            elif isinstance(val, SerializableArray):
                return val.to_json_list(check_validity=check_validity, strict=strict)
            elif isinstance(val, ParametersCollection):
                return val.to_dict()
            elif isinstance(val, int) or isinstance(val, str) or isinstance(val, float):
                return val
            elif isinstance(val, numpy.datetime64):
                out2 = str(val)
                return out2 + 'Z' if out2[-1] != 'Z' else out2
            elif isinstance(val, complex):
                return {'Real': val.real, 'Imag': val.imag}
            elif isinstance(val, date):  # probably never present
                return val.isoformat()
            elif isinstance(val, datetime):  # probably never present
                return val.isoformat(sep='T')
            else:
                raise ValueError(
                    'a entry for class {} using tag {} is of type {}, and serialization has not '
                    'been implemented'.format(self.__class__.__name__, field, type(val)))

        if check_validity:
            if not self.is_valid(stack=False):
                msg = "{} is not valid,\n\t" \
                      "and cannot be SAFELY serialized to a dictionary valid in " \
                      "the SICD standard.".format(self.__class__.__name__)
                if strict:
                    raise ValueError(msg)
                logger.warning(msg)

        out = OrderedDict()

        for attribute in self._fields:
            if attribute in exclude:
                continue

            value = getattr(self, attribute)
            if value is None:
                continue

            if isinstance(value, (numpy.ndarray, list)):
                array_tag = self._collections_tags.get(attribute, None)
                if array_tag is None:
                    raise AttributeError(
                        'The value associated with attribute {} in an instance of class {} is of type {}, '
                        'but nothing is populated in the _collection_tags dictionary.'.format(
                            attribute, self.__class__.__name__, type(value)))
                child_tag = array_tag.get('child_tag', None)
                if child_tag is None:
                    raise AttributeError(
                        'The value associated with attribute {} in an instance of class {} is of type {}, '
                        'but `child_tag` is not populated in the _collection_tags dictionary.'.format(
                            attribute, self.__class__.__name__, type(value)))
                if isinstance(value, numpy.ndarray):
                    out[attribute] = serialize_array(child_tag, value)
                else:
                    out[attribute] = serialize_list(child_tag, value)
            else:
                out[attribute] = serialize_plain(attribute, value)
        return out

    def copy(self):
        """
        Create a deep copy.

        Returns
        -------

        """

        return self.__class__.from_dict(copy.deepcopy(self.to_dict(check_validity=False)))

    def to_xml_bytes(self, urn=None, tag=None, check_validity=False, strict=DEFAULT_STRICT):
        """
        Gets a bytes array, which corresponds to the xml string in utf-8 encoding,
        identified as using the namespace given by `urn` (if given).

        Parameters
        ----------
        urn : None|str|dict
            The xml namespace string or dictionary describing the xml namespace.
        tag : None|str
            The root node tag to use. If not given, then the class name will be used.
        check_validity : bool
            Check whether the element is valid before serializing, by calling :func:`is_valid`.
        strict : bool
            Only used if `check_validity = True`. In that case, if `True` then raise an
            Exception (of appropriate type) if the structure is not valid, if `False` then log a
            hopefully helpful message.

        Returns
        -------
        bytes
            bytes array from :func:`ElementTree.tostring()` call.
        """
        if tag is None:
            tag = self.__class__.__name__
        the_etree = ElementTree.ElementTree()
        node = self.to_node(the_etree, tag, ns_key=getattr(self, '_xml_ns_key', None),
                            check_validity=check_validity, strict=strict)

        if urn is None:
            pass
        elif isinstance(urn, str):
            node.attrib['xmlns'] = urn
        elif isinstance(urn, dict):
            for key in urn:
                node.attrib[key] = urn[key]
        else:
            raise TypeError('Expected string or dictionary of string for urn, got type {}'.format(type(urn)))
        return ElementTree.tostring(node, encoding='utf-8', method='xml')

    def to_xml_string(self, urn=None, tag=None, check_validity=False, strict=DEFAULT_STRICT):
        """
        Gets a xml string with utf-8 encoding, identified as using the namespace
        given by `urn` (if given).

        Parameters
        ----------
        urn : None|str|dict
            The xml namespace or dictionary describing the xml namespace.
        tag : None|str
            The root node tag to use. If not given, then the class name will be used.
        check_validity : bool
            Check whether the element is valid before serializing, by calling :func:`is_valid`.
        strict : bool
            Only used if `check_validity = True`. In that case, if `True` then raise an
            Exception (of appropriate type) if the structure is not valid, if `False` then log a
            hopefully helpful message.

        Returns
        -------
        str
            xml string from :func:`ElementTree.tostring()` call.
        """

        return self.to_xml_bytes(urn=urn, tag=tag, check_validity=check_validity, strict=strict).decode('utf-8')


class Arrayable(object):
    """Abstract class specifying basic functionality for assigning from/to an array"""

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type object.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple

        Returns
        -------
        Arrayable
        """

        raise NotImplementedError

    def get_array(self, dtype=numpy.float64):
        """Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
        """
        raise NotImplementedError

    def __getitem__(self, item):
        return self.get_array()[item]


class SerializableArray(object):
    __slots__ = (
        '_child_tag', '_child_type', '_array', '_name', '_minimum_length',
        '_maximum_length', '_xml_ns', '_xml_ns_key')
    _default_minimum_length = 0
    _default_maximum_length = 2**32
    _set_size = True
    _size_var_name = 'size'
    _set_index = True
    _index_var_name = 'index'

    def __init__(self, coords=None, name=None, child_tag=None, child_type=None,
                 minimum_length=None, maximum_length=None, _xml_ns=None, _xml_ns_key=None):
        self._xml_ns = _xml_ns
        self._xml_ns_key = _xml_ns_key
        self._array = None
        if name is None:
            raise ValueError('The name parameter is required.')
        if not isinstance(name, str):
            raise TypeError(
                'The name parameter is required to be an instance of str, got {}'.format(type(name)))
        self._name = name

        if child_tag is None:
            raise ValueError('The child_tag parameter is required.')
        if not isinstance(child_tag, str):
            raise TypeError(
                'The child_tag parameter is required to be an instance of str, got {}'.format(type(child_tag)))
        self._child_tag = child_tag

        if child_type is None:
            raise ValueError('The child_type parameter is required.')
        if not issubclass(child_type, Serializable):
            raise TypeError('The child_type is required to be a subclass of Serializable.')
        self._child_type = child_type

        if minimum_length is None:
            self._minimum_length = self._default_minimum_length
        else:
            self._minimum_length = max(int(minimum_length), 0)

        if maximum_length is None:
            self._maximum_length = max(self._default_maximum_length, self._minimum_length)
        else:
            self._maximum_length = max(int(maximum_length), self._minimum_length)

        self.set_array(coords)

    def __len__(self):
        return self.size

    def __getitem__(self, index):
        return self._array[index]

    def __setitem__(self, index, value):
        if value is None:
            raise TypeError('Elements of {} must be of type {}, not None'.format(self._name, self._child_type))
        self._array[index] = parse_serializable(value, self._name, self, self._child_type)

    def log_validity_error(self, msg):
        """
        Log a validation error message.

        Parameters
        ----------
        msg : str
        """

        valid_logger.error('{}:{} {}'.format(self.__class__.__name__, self._name, msg))

    def log_validity_warning(self, msg):
        """
        Log a validation warning message.

        Parameters
        ----------
        msg : str
        """

        valid_logger.warning('{}:{} {}'.format(self.__class__.__name__, self._name, msg))

    def log_validity_info(self, msg):
        """
        Log a validation info message.

        Parameters
        ----------
        msg : str
        """

        valid_logger.info('{}:{} {}'.format(self.__class__.__name__, self._name, msg))

    def is_valid(self, recursive=False, stack=False):
        """Returns the validity of this object according to the schema. This is done by inspecting that the
        array is populated.

        Parameters
        ----------
        recursive : bool
            True if we recursively check that children are also valid. This may result in verbose (i.e. noisy) logging.
        stack : bool
            Should we print error messages recursively, for a stack type situation?
        Returns
        -------
        bool
            condition for validity of this element
        """
        if self._array is None:
            self.log_validity_error("Unpopulated array")
            return False
        if not recursive:
            return True
        valid_children = True
        for i, entry in enumerate(self._array):
            good = entry.is_valid(recursive=True, stack=stack)
            if not good and stack:
                self.log_validity_error("Issue discovered with entry {}".format(i))
            valid_children &= good
        return valid_children

    @property
    def size(self):  # type: () -> int
        """
        int: the size of the array.
        """

        if self._array is None:
            return 0
        else:
            return self._array.size

    def get_array(self, dtype='object', **kwargs):
        """Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return.
        kwargs : keyword arguments for calls of the form child.get_array(**kwargs)

        Returns
        -------
        numpy.ndarray
            * If `dtype` == 'object'`, then the literal array of
              child objects is returned. *Note: Beware of mutating the elements.*
            * If `dtype` has any other value, then the return value will be tried
              as `numpy.array([child.get_array(dtype=dtype, **kwargs) for child in array]`.
            * If there is any error, then `None` is returned.
        """

        if dtype in ['object', numpy.dtype('object')]:
            return self._array
        else:
            # noinspection PyBroadException
            try:
                return numpy.array(
                    [child.get_array(dtype=dtype, **kwargs) for child in self._array], dtype=dtype)
            except Exception:
                return None

    def set_array(self, coords):
        """
        Sets the underlying array.

        Parameters
        ----------
        coords : numpy.ndarray|list|tuple

        Returns
        -------
        None
        """

        if coords is None:
            self._array = None
            return
        array = parse_serializable_array(
            coords, 'coords', self, self._child_type, self._child_tag)
        if not (self._minimum_length <= array.size <= self._maximum_length):
            raise ValueError(
                'Field {} is required to be an array with {} <= length <= {}, and input of length {} '
                'was received'.format(self._name, self._minimum_length, self._maximum_length, array.size))

        self._array = array
        self._check_indices()

    def _check_indices(self):
        if not self._set_index:
            return
        for i, entry in enumerate(self._array):
            try:
                setattr(entry, self._index_var_name, i+1)
            except (AttributeError, ValueError, TypeError):
                continue

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT):
        if self.size == 0:
            return None  # nothing to be done
        if ns_key is None:
            anode = create_new_node(doc, tag, parent=parent)
        else:
            anode = create_new_node(doc, '{}:{}'.format(ns_key, tag), parent=parent)
        if self._set_size:
            anode.attrib[self._size_var_name] = str(self.size)
        for i, entry in enumerate(self._array):
            entry.to_node(doc, self._child_tag, ns_key=ns_key, parent=anode,
                          check_validity=check_validity, strict=strict)
        return anode

    @classmethod
    def from_node(cls, node, name, child_tag, child_type, **kwargs):
        return cls(coords=node, name=name, child_tag=child_tag, child_type=child_type, **kwargs)

    def to_json_list(self, check_validity=False, strict=DEFAULT_STRICT):
        """
        For json serialization.
        Parameters
        ----------
        check_validity : bool
            passed through to child_type.to_dict() method.
        strict : bool
            passed through to child_type.to_dict() method.

        Returns
        -------
        List[dict]
        """

        if self.size == 0:
            return []
        return [entry.to_dict(check_validity=check_validity, strict=strict) for entry in self._array]


class ParametersCollection(object):
    __slots__ = ('_name', '_child_tag', '_dict', '_xml_ns', '_xml_ns_key')

    def __init__(self, collection=None, name=None, child_tag='Parameters', _xml_ns=None, _xml_ns_key=None):
        self._dict = None
        self._xml_ns = _xml_ns
        self._xml_ns_key = _xml_ns_key
        if name is None:
            raise ValueError('The name parameter is required.')
        if not isinstance(name, str):
            raise TypeError(
                'The name parameter is required to be an instance of str, got {}'.format(type(name)))
        self._name = name

        if child_tag is None:
            raise ValueError('The child_tag parameter is required.')
        if not isinstance(child_tag, str):
            raise TypeError(
                'The child_tag parameter is required to be an instance of str, got {}'.format(type(child_tag)))
        self._child_tag = child_tag

        self.set_collection(collection)

    def __delitem__(self, key):
        if self._dict is not None:
            del self._dict[key]

    def __getitem__(self, key):
        if self._dict is not None:
            return self._dict[key]
        raise KeyError('Dictionary does not contain key {}'.format(key))

    def __setitem__(self, name, value):
        if not isinstance(name, str):
            raise ValueError('Parameter name must be of type str, got {}'.format(type(name)))
        if not isinstance(value, str):
            raise ValueError('Parameter name must be of type str, got {}'.format(type(value)))

        if self._dict is None:
            self._dict = OrderedDict()
        self._dict[name] = value

    def get(self, key, default=None):
        if self._dict is not None:
            return self._dict.get(key, default)
        return default

    def set_collection(self, value):
        if value is None:
            self._dict = None
        else:
            self._dict = parse_parameters_collection(value, self._name, self)

    def get_collection(self):
        return self._dict

    # noinspection PyUnusedLocal
    def to_node(self, doc, ns_key=None, parent=None, check_validity=False, strict=False):
        if self._dict is None:
            return None  # nothing to be done
        for name in self._dict:
            value = self._dict[name]
            if not isinstance(value, str):
                value = str(value)
            if ns_key is None:
                node = create_text_node(doc, self._child_tag, value, parent=parent)
            else:
                node = create_text_node(doc, '{}:{}'.format(ns_key, self._child_tag), value, parent=parent)
            node.attrib['name'] = name

    # noinspection PyUnusedLocal
    def to_dict(self, check_validity=False, strict=False):
        return copy.deepcopy(self._dict)



"""
Definition for the ImageInfo NGA modified RDE/AFRL labeling object
"""

__classification__ = "UNCLASSIFIED"
__authors__ = "Thomas McCullough"

from typing import Optional
import numpy

from sarpy.io.xml.base import Serializable, Arrayable
from sarpy.io.xml.descriptors import StringDescriptor, SerializableDescriptor, \
    IntegerDescriptor, StringEnumDescriptor, DateTimeDescriptor, FloatDescriptor, \
    BooleanDescriptor
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.blocks import LatLonType

from .base import DEFAULT_STRICT
from .blocks import RangeCrossRangeType, ProjectionPerturbationType


class NumPixelsType(Serializable, Arrayable):
    """A row and column attribute container - used as indices into array(s)."""
    _fields = ('NumRows', 'NumCols')
    _required = _fields
    NumRows = IntegerDescriptor(
        'NumRows', _required, strict=True, docstring='The number of rows.')  # type: int
    NumCols = IntegerDescriptor(
        'NumCols', _required, strict=True, docstring='The number of columns.')  # type: int

    def __init__(self, NumRows=None, NumCols=None, **kwargs):
        """
        Parameters
        ----------
        NumRows : int
        NumCols : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.NumRows, self.NumCols = NumRows, NumCols
        super(NumPixelsType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.int64):
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [NumRows, NumCols]
        """

        return numpy.array([self.NumRows, self.NumCols], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [NumRows, NumCols]

        Returns
        -------
        NumPixelsType
        """
        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError('Expected array to be of length 2, and received {}'.format(array))
            return cls(NumRows=array[0], NumCols=array[1])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class ClassificationMarkingsType(Serializable):
    _fields = (
        'Classification', 'Restrictions', 'ClassifiedBy', 'DeclassifyOn', 'DerivedFrom')
    _required = ('Classification', 'Restrictions')
    # descriptors
    Classification = StringDescriptor(
        'Classification', _required, default_value='',
        docstring='The image classification')  # type: str
    Restrictions = StringDescriptor(
        'Restrictions', _required, default_value='',
        docstring='Additional caveats to the classification')  # type: str
    ClassifiedBy = StringDescriptor(
        'ClassifiedBy', _required)  # type: Optional[str]
    DeclassifyOn = StringDescriptor(
        'DeclassifyOn', _required)  # type: Optional[str]
    DerivedFrom = StringDescriptor(
        'DerivedFrom', _required)  # type: Optional[str]

    def __init__(self, Classification='', Restrictions='', ClassifiedBy=None,
                 DeclassifyOn=None, DerivedFrom=None, **kwargs):
        """
        Parameters
        ----------
        Classification : str
        Restrictions : str
        ClassifiedBy : None|str
        DeclassifyOn : None|str
        DerivedFrom : None|str
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Classification = Classification
        self.Restrictions = Restrictions
        self.ClassifiedBy = ClassifiedBy
        self.DeclassifyOn = DeclassifyOn
        self.DerivedFrom = DerivedFrom
        super(ClassificationMarkingsType, self).__init__(**kwargs)


class StringRangeCrossRangeType(Serializable):
    """
    A range and cross range attribute container
    """
    _fields = ('Range', 'CrossRange')
    _required = _fields
    # descriptors
    Range = StringDescriptor(
        'Range', _required, strict=True, docstring='The Range attribute.')  # type: str
    CrossRange = StringDescriptor(
        'CrossRange', _required, strict=True, docstring='The Cross Range attribute.')  # type: str

    def __init__(self, Range=None, CrossRange=None, **kwargs):
        """
        Parameters
        ----------
        Range : str
        CrossRange : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Range, self.CrossRange = Range, CrossRange
        super(StringRangeCrossRangeType, self).__init__(**kwargs)


class ImageCornerType(Serializable):
    _fields = (
        'UpperLeft', 'UpperRight', 'LowerRight', 'LowerLeft')
    _required = _fields
    # descriptors
    UpperLeft = SerializableDescriptor(
        'UpperLeft', LatLonType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: LatLonType
    UpperRight = SerializableDescriptor(
        'UpperRight', LatLonType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: LatLonType
    LowerRight = SerializableDescriptor(
        'LowerRight', LatLonType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: LatLonType
    LowerLeft = SerializableDescriptor(
        'LowerLeft', LatLonType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: LatLonType

    def __init__(self, UpperLeft=None, UpperRight=None,
                 LowerRight=None, LowerLeft=None, **kwargs):
        """
        Parameters
        ----------
        UpperLeft : LatLonType|numpy.ndarray|list|tuple
        UpperRight : LatLonType|numpy.ndarray|list|tuple
        LowerRight : LatLonType|numpy.ndarray|list|tuple
        LowerLeft : LatLonType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.UpperLeft = UpperLeft
        self.UpperRight = UpperRight
        self.LowerRight = LowerRight
        self.LowerLeft = LowerLeft
        super(ImageCornerType, self).__init__(**kwargs)


class PixelSpacingType(Serializable):
    _fields = ('PixelSpacing', )
    _required = _fields
    # descriptors
    PixelSpacing = SerializableDescriptor(
        'PixelSpacing', RangeCrossRangeType, _required, strict=DEFAULT_STRICT,
        docstring='The center-to-center pixel spacing in meters.')  # type: RangeCrossRangeType

    def __init__(self, PixelSpacing=None, **kwargs):
        """
        Parameters
        ----------
        PixelSpacing : RangeCrossRangeType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PixelSpacing = PixelSpacing
        super(PixelSpacingType, self).__init__(**kwargs)


class ImageInfoType(Serializable):
    _fields = (
        'DataFilename', 'ClassificationMarkings', 'Filetype', 'DataCheckSum',
        'DataSize', 'DataPlane', 'DataDomain', 'DataType', 'BitsPerSample',
        'DataFormat', 'DataByteOrder', 'NumPixels', 'ImageCollectionDate', 'ZuluOffset',
        'SensorReferencePoint', 'SensorCalibrationFactor', 'DataCalibrated',
        'Resolution', 'PixelSpacing', 'WeightingType', 'OverSamplingFactor',
        'IPRWidth3dB', 'ImageQualityDescription', 'ImageHeading',
        'ImageCorners', 'SlantPlane', 'GroundPlane', 'SceneCenterReferenceLine', 
        'ProjectionPerturbation')
    _required = (
        'DataFilename', 'ClassificationMarkings', 'DataCheckSum', 'DataPlane',
        'DataDomain', 'DataType', 'DataFormat', 'NumPixels', 'ImageCollectionDate',
        'SensorReferencePoint', 'DataCalibrated', 'Resolution', 'PixelSpacing',
        'WeightingType', 'ImageCorners')
    _numeric_format = {
        'ImageHeading': '0.17G', 'SensorCalibrationFactor': '0.17G',
        'SceneCenterReferenceLine': '0.17G', }
    # descriptors
    DataFilename = StringDescriptor(
        'DataFilename', _required,
        docstring='The base file name to which this information pertains')  # type: str
    ClassificationMarkings = SerializableDescriptor(
        'ClassificationMarkings', ClassificationMarkingsType, _required,
        docstring='The classification information')  # type: ClassificationMarkingsType
    Filetype = StringDescriptor(
        'Filetype', _required,
        docstring='The image file type')  # type: Optional[str]
    DataCheckSum = StringDescriptor(
        'DataCheckSum', _required,
        docstring='The 32 character (hexidecimal digest) MD5 checksum of the '
                  'full image file')  # type: str
    DataSize = IntegerDescriptor(
        'DataSize', _required,
        docstring='The image size in bytes')  # type: Optional[int]
    DataPlane = StringEnumDescriptor(
        'DataPlane', {'Slant', 'Ground'}, _required, default_value='Slant',
        docstring='The image plane.')  # type: str
    DataDomain = StringEnumDescriptor(
        'DataDomain', {'Image', }, _required,    # todo: values
        docstring='The image data domain')  # type: str
    DataType = StringEnumDescriptor(
        'DataType', {'Magnitude/Phase', 'In-phase/Quadrature'}, _required,
        docstring='The image data type')  # type: str
    BitsPerSample = IntegerDescriptor(
        'BitsPerSample', _required,
        docstring='The number of bits per sample')  # type: Optional[int]
    DataFormat = StringDescriptor(
        'DataFormat', _required,
        docstring='The image data format')  # type: str
    DataByteOrder = StringEnumDescriptor(
        'DataByteOrder', {'Big-Endian', 'Little-Endian'}, _required,
        docstring='The image data byte order.')  # type: Optional[str]
    NumPixels = SerializableDescriptor(
        'NumPixels', NumPixelsType, _required,
        docstring='The number of image pixels')  # type: NumPixelsType
    ImageCollectionDate = DateTimeDescriptor(
        'ImageCollectionDate', _required,
        docstring='The date/time of the image collection in UTC')  # type: Optional[numpy.datetime64]
    ZuluOffset = StringDescriptor(
        'ZuluOffset', _required,
        docstring='The local time offset from UTC')  # type: Optional[str]
    SensorReferencePoint = StringEnumDescriptor(
        'DataPlane', {'Left', 'Right', 'Top', 'Bottom'}, _required,
        docstring='Description of the sensor location relative to the scene.')  # type: Optional[str]
    SensorCalibrationFactor = FloatDescriptor(
        'SensorCalibrationFactor', _required,
        docstring='Multiplicative factor used to scale raw image data to the return '
                  'of a calibrated reference reflector or active source')  # type: Optional[float]
    DataCalibrated = BooleanDescriptor(
        'DataCalibrated', _required,
        docstring='Has the data been calibrated?')  # type: bool
    Resolution = SerializableDescriptor(
        'Resolution', RangeCrossRangeType, _required,
        docstring='Resolution (intrinsic) of the sensor system/mode in meters.')  # type: RangeCrossRangeType
    PixelSpacing = SerializableDescriptor(
        'PixelSpacing', RangeCrossRangeType, _required,
        docstring='Pixel spacing of the image in meters.')  # type: RangeCrossRangeType
    WeightingType = SerializableDescriptor(
        'WeightingType', StringRangeCrossRangeType, _required,
        docstring='Weighting function applied to the image during formation.')  # type: StringRangeCrossRangeType
    OverSamplingFactor = SerializableDescriptor(
        'OverSamplingFactor', RangeCrossRangeType, _required,
        docstring='The factor by which the pixel space is oversampled.')  # type: Optional[RangeCrossRangeType]
    IPRWidth3dB = SerializableDescriptor(
        'IPRWidth3dB', RangeCrossRangeType, _required,
        docstring='The 3 dB system impulse response with, in meters')  # type: Optional[RangeCrossRangeType]
    ImageQualityDescription = StringDescriptor(
        'ImageQualityDescription', _required,
        docstring='General description of image quality')  # type: Optional[str]
    ImageHeading = FloatDescriptor(
        'ImageHeading', _required,
        docstring='Image heading relative to True North, in degrees')  # type: Optional[float]
    ImageCorners = SerializableDescriptor(
        'ImageCorners', ImageCornerType, _required,
        docstring='The image corners')  # type: ImageCornerType
    SlantPlane = SerializableDescriptor(
        'SlantPlane', PixelSpacingType, _required,
        docstring='The slant plane pixel spacing')  # type: Optional[PixelSpacingType]
    GroundPlane = SerializableDescriptor(
        'GroundPlane', PixelSpacingType, _required,
        docstring='The ground plane pixel spacing')  # type: Optional[PixelSpacingType]
    SceneCenterReferenceLine = FloatDescriptor(
        'SceneCenterReferenceLine', _required,
        docstring='The ideal line (heading) at the intersection of the radar '
                  'line-of-sight with the horizontal reference plane '
                  'created by the forward motion of the aircraft, '
                  'in degrees')  # type: Optional[float]
    ProjectionPerturbation = SerializableDescriptor(
        'ProjectionPerturbation', ProjectionPerturbationType, _required, 
        docstring='')  # type: Optional[ProjectionPerturbationType]

    def __init__(self, DataFilename=None, ClassificationMarkings=None,
                 FileType=None, DataCheckSum=None, DataSize=None,
                 DataPlane='Slant', DataDomain=None, DataType=None,
                 BitsPerSample=None, DataFormat=None, DataByteOrder=None, NumPixels=None,
                 ImageCollectionDate=None, ZuluOffset=None,
                 SensorReferencePoint=None, SensorCalibrationFactor=None,
                 DataCalibrated=None, Resolution=None, PixelSpacing=None,
                 WeightingType=None, OverSamplingFactor=None, IPRWidth3dB=None,
                 ImageQualityDescription=None, ImageHeading=None, ImageCorners=None,
                 SlantPlane=None, GroundPlane=None, SceneCenterReferenceLine=None,
                 ProjectionPerturbation=None,
                 **kwargs):
        """
        Parameters
        ----------
        DataFilename : str
        ClassificationMarkings : ClassificationMarkingsType
        FileType : str
        DataCheckSum : str
        DataSize : int
        DataPlane : str
        DataDomain : None|str
        DataType : None|str
        BitsPerSample : None|int
        DataFormat : None|str
        DataByteOrder : None|str
        NumPixels : NumPixelsType|numpy.ndarray|list|tuple
        ImageCollectionDate : numpy.datetime64|datetime|date|str
        ZuluOffset : None|str
        SensorReferencePoint : None|str
        SensorCalibrationFactor : None|float
        DataCalibrated : bool
        Resolution : RangeCrossRangeType|numpy.ndarray|list|tuple
        PixelSpacing : RangeCrossRangeType|numpy.ndarray|list|tuple
        WeightingType : StringRangeCrossRangeType
        OverSamplingFactor : None|RangeCrossRangeType
        IPRWidth3dB : None|RangeCrossRangeType|numpy.ndarray|list|tuple
        ImageQualityDescription : None|str
        ImageHeading : None|float
        ImageCorners : ImageCornerType
        SlantPlane : None|PixelSpacingType
        GroundPlane : None|PixelSpacingType
        SceneCenterReferenceLine : None|float
        ProjectionPerturbation : None|ProjectionPerturbationType
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']

        self.DataFilename = DataFilename

        if ClassificationMarkings is None:
            self.ClassificationMarkings = ClassificationMarkingsType()
        else:
            self.ClassificationMarkings = ClassificationMarkings

        self.Filetype = FileType
        self.DataCheckSum = DataCheckSum
        self.DataSize = DataSize
        self.DataPlane = DataPlane
        self.DataDomain = DataDomain
        self.DataType = DataType

        self.BitsPerSample = BitsPerSample
        self.DataFormat = DataFormat
        self.DataByteOrder = DataByteOrder
        self.NumPixels = NumPixels
        self.ImageCollectionDate = ImageCollectionDate
        self.ZuluOffset = ZuluOffset

        self.SensorReferencePoint = SensorReferencePoint
        self.SensorCalibrationFactor = SensorCalibrationFactor
        self.DataCalibrated = DataCalibrated
        self.Resolution = Resolution
        self.PixelSpacing = PixelSpacing
        self.WeightingType = WeightingType
        self.OverSamplingFactor = OverSamplingFactor
        self.IPRWidth3dB = IPRWidth3dB

        self.ImageQualityDescription = ImageQualityDescription
        self.ImageHeading = ImageHeading
        self.ImageCorners = ImageCorners
        self.SlantPlane = SlantPlane
        self.GroundPlane = GroundPlane
        self.SceneCenterReferenceLine = SceneCenterReferenceLine
        self.ProjectionPerturbation = ProjectionPerturbation
        super(ImageInfoType, self).__init__(**kwargs)

    @classmethod
    def from_sicd(cls, sicd, base_file_name, file_type='NITF02.10', md5_checksum=None):
        """
        Construct the ImageInfo from the sicd object and given image file name.

        Parameters
        ----------
        sicd : SICDType
        base_file_name : str
        file_type : str
            The file type. This should probably always be NITF02.10 for now.
        md5_checksum : None|str
            The md5 checksum of the full image file.

        Returns
        -------
        ImageInfoType
        """

        pixel_type = sicd.ImageData.PixelType
        if pixel_type == 'RE32F_IM32F':
            data_type = 'In-phase/Quadrature'
            bits_per_sample = 32
            data_format = 'float'
        elif pixel_type == 'RE16I_IM16I':
            data_type = 'In-phase/Quadrature'
            bits_per_sample = 16
            data_format = 'integer'
        elif pixel_type == 'AMP8I_PHS8I':
            data_type = 'Magnitude/Phase'
            bits_per_sample = 8
            data_format = 'unsigned integer'
        else:
            raise ValueError('Unhandled')

        data_cal = sicd.Radiometric is not None

        icps = ImageCornerType(
            UpperLeft=sicd.GeoData.ImageCorners.FRFC,
            UpperRight=sicd.GeoData.ImageCorners.FRLC,
            LowerRight=sicd.GeoData.ImageCorners.LRLC,
            LowerLeft=sicd.GeoData.ImageCorners.LRFC)

        if sicd.Grid.ImagePlane == 'SLANT':
            data_plane = 'Slant'
        elif sicd.Grid.ImagePlane == 'Ground':
            data_plane = 'Ground'
        else:
            data_plane = None

        has_perturb = False
        proj_perturb = None
        coa = sicd.coa_projection
        if coa is not None:
            delta_arp = coa.delta_arp
            if numpy.any(delta_arp != 0):
                has_perturb = True
            else:
                delta_arp = None

            delta_varp = coa.delta_varp
            if numpy.any(delta_varp != 0):
                has_perturb = True
            else:
                delta_varp = None

            delta_range = coa.delta_range
            if delta_range != 0:
                has_perturb = True
            else:
                delta_range = None
            
            if has_perturb:
                proj_perturb = ProjectionPerturbationType(
                    CoordinateFrame='ECF', 
                    DeltaArp=delta_arp, 
                    DeltaVarp=delta_varp, 
                    DeltaRange=delta_range)

        return ImageInfoType(
            DataFilename=base_file_name,
            ClassificationMarkings=ClassificationMarkingsType(
                Classification=sicd.CollectionInfo.Classification),
            FileType=file_type,
            DataCheckSum=md5_checksum,
            DataPlane=data_plane,
            DataType=data_type,
            DataCalibrated=data_cal,
            BitsPerSample=bits_per_sample,
            DataDomain='Image',
            DataFormat=data_format,
            DataByteOrder='Big-Endian',
            NumPixels=(sicd.ImageData.NumRows, sicd.ImageData.NumCols),
            ImageCollectionDate=sicd.Timeline.CollectStart,
            SensorReferencePoint='Top',
            Resolution=(sicd.Grid.Row.ImpRespWid, sicd.Grid.Col.ImpRespWid),
            PixelSpacing=(sicd.Grid.Row.SS, sicd.Grid.Col.SS),
            WeightingType=StringRangeCrossRangeType(
                Range=sicd.Grid.Row.WgtType.WindowName,
                CrossRange=sicd.Grid.Col.WgtType.WindowName),
            ImageHeading=sicd.SCPCOA.AzimAng,
            ImageCorners=icps, 
            ProjectionPerturbation=proj_perturb)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class IOMAPA_6Type(TREElement):
    def __init__(self, value):
        super(IOMAPA_6Type, self).__init__()
        self.add_field('BAND_NUMBER', 's', 3, value)
        self.add_field('MAP_SELECT', 's', 1, value)
        self.add_field('S2', 's', 2, value)


class IOMAPA_6(TREExtension):
    _tag_value = 'IOMAPA'
    _data_type = IOMAPA_6Type


class IOMAPA_16Type(TREElement):
    def __init__(self, value):
        super(IOMAPA_16Type, self).__init__()
        self.add_field('BAND_NUMBER', 's', 3, value)
        self.add_field('MAP_SELECT', 's', 1, value)
        self.add_field('TABLE_ID', 's', 2, value)
        self.add_field('S1', 's', 2, value)
        self.add_field('S2', 's', 2, value)
        self.add_field('R_WHOLE', 's', 3, value)
        self.add_field('R_FRACTION', 's', 3, value)


class IOMAPA_16(TREExtension):
    _tag_value = 'IOMAPA'
    _data_type = IOMAPA_16Type


class SEGMENT(TREElement):
    def __init__(self, value):
        super(SEGMENT, self).__init__()
        self.add_field('OUT_B0', 'b', 4, value)
        self.add_field('OUT_B1', 'b', 4, value)
        self.add_field('OUT_B2', 'b', 4, value)
        self.add_field('OUT_B3', 'b', 4, value)
        self.add_field('OUT_B4', 'b', 4, value)
        self.add_field('OUT_B5', 'b', 4, value)


class IOMAPA_91Type(TREElement):
    def __init__(self, value):
        super(IOMAPA_91Type, self).__init__()
        self.add_field('BAND_NUMBER', 's', 3, value)
        self.add_field('MAP_SELECT', 's', 1, value)
        self.add_field('TABLE_ID', 's', 1, value)
        self.add_field('S1', 's', 2, value)
        self.add_field('S2', 's', 2, value)
        self.add_field('NO_OF_SEGMENTS', 'd', 1, value)
        self.add_field('XOB_1', 's', 4, value)
        self.add_field('XOB_2', 's', 4, value)
        self.add_loop('SEGMENTs', self.NO_OF_SEGMENTS, SEGMENT, value)


class IOMAPA_91(TREExtension):
    _tag_value = 'IOMAPA'
    _data_type = IOMAPA_91Type


class MAP(TREElement):
    def __init__(self, value):
        super(MAP, self).__init__()
        self.add_field('OUTPUT_MAP_VALUE', 's', 2, value)


class IOMAPA_8202Type(TREElement):
    def __init__(self, value):
        super(IOMAPA_8202Type, self).__init__()
        self.add_field('BAND_NUMBER', 's', 3, value)
        self.add_field('MAP_SELECT', 's', 1, value)
        self.add_field('TABLE_ID', 's', 2, value)
        self.add_field('S1', 's', 2, value)
        self.add_field('S2', 's', 2, value)
        self.add_loop('MAPs', 4096, MAP, value)


class IOMAPA_8202(TREExtension):
    _tag_value = 'IOMAPA'
    _data_type = IOMAPA_8202Type


class IOMAPA(TREExtension):
    _tag_value = 'IOMAPA'

    def __init__(self):
        raise ValueError(
            'Not to be implemented directly. '
            'Use of one IOMAPA_6, IOMAPA_16, IOMAPA_91, or IOMAPA_8202')

    @classmethod
    def from_bytes(cls, value, start):
        """

        Parameters
        ----------
        value : bytes
        start : int

        Returns
        -------
        IOMAPA_6|IOMAPA_16|IOMAPA_91|IOMAPA_8202
        """

        tag_value = value[start:start+6].decode('utf-8').strip()
        if tag_value != cls._tag_value:
            raise ValueError('tag value must be {}. Got {}'.format(cls._tag_value, tag_value))

        lng = int(value[start+6:start+11])
        if lng == 6:
            return IOMAPA_6.from_bytes(value, start)
        elif lng == 16:
            return IOMAPA_16.from_bytes(value, start)
        elif lng == 91:
            return IOMAPA_91.from_bytes(value, start)
        elif lng == 8202:
            return IOMAPA_8202.from_bytes(value, start)
        else:
            raise ValueError('the data must be length 6, 16, 91, or 8202. Got {}'.format(lng))




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PIAEVAType(TREElement):
    def __init__(self, value):
        super(PIAEVAType, self).__init__()
        self.add_field('EVENTNAME', 's', 38, value)
        self.add_field('EVENTTYPE', 's', 8, value)


class PIAEVA(TREExtension):
    _tag_value = 'PIAEVA'
    _data_type = PIAEVAType



"""
The Dwell parameters definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import List

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import StringDescriptor, SerializableDescriptor, SerializableListDescriptor
from sarpy.io.complex.sicd_elements.blocks import Poly2DType

from .base import DEFAULT_STRICT


class CODTimeType(Serializable):
    """
    Center of Dwell (COD) time polynomial object.
    """

    _fields = ('Identifier', 'CODTimePoly')
    _required = _fields
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='String that uniquely identifies this COD Time '
                  'polynomial.')  # type: str
    CODTimePoly = SerializableDescriptor(
        'CODTimePoly', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='The polynomial.')  # type: Poly2DType

    def __init__(self, Identifier=None, CODTimePoly=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        CODTimePoly : Poly2DType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.CODTimePoly = CODTimePoly
        super(CODTimeType, self).__init__(**kwargs)


class DwellTimeType(Serializable):
    """
    The dwell time polynomial object.
    """
    _fields = ('Identifier', 'DwellTimePoly')
    _required = _fields
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='String that uniquely identifies this Dwell Time '
                  'polynomial.')  # type: str
    DwellTimePoly = SerializableDescriptor(
        'DwellTimePoly', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='The polynomial.')  # type: Poly2DType

    def __init__(self, Identifier=None, DwellTimePoly=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        DwellTimePoly : Poly2DType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.DwellTimePoly = DwellTimePoly
        super(DwellTimeType, self).__init__(**kwargs)


class DwellType(Serializable):
    """
    Parameters that specify the dwell time supported by the signal arrays
    contained in the CPHD product.
    """

    _fields = ('NumCODTimes', 'CODTimes', 'NumDwellTimes', 'DwellTimes')
    _required = _fields
    _collections_tags = {
        'CODTimes': {'array': False, 'child_tag': 'CODTime'},
        'DwellTimes': {'array': False, 'child_tag': 'DwellTime'}}
    # descriptors
    CODTimes = SerializableListDescriptor(
        'CODTimes', CODTimeType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='The Center of Dwell (COD) time polynomials.')  # type: List[CODTimeType]
    DwellTimes = SerializableListDescriptor(
        'DwellTimes', DwellTimeType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='The dwell time polynomials.')  # type: List[DwellTimeType]

    def __init__(self, CODTimes=None, DwellTimes=None, **kwargs):
        """

        Parameters
        ----------
        CODTimes : List[CODTimeType]
        DwellTimes : List[DwellTimeType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CODTimes = CODTimes
        self.DwellTimes = DwellTimes
        super(DwellType, self).__init__(**kwargs)

    @property
    def NumCODTimes(self):
        """
        int: The number of cod time polynomial elements.
        """

        if self.CODTimes is None:
            return 0
        else:
            return len(self.CODTimes)

    @property
    def NumDwellTimes(self):
        """
        int: The number of dwell time polynomial elements.
        """

        if self.DwellTimes is None:
            return 0
        else:
            return len(self.DwellTimes)



"""
The SceneCoordinates type definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, List

from sarpy.io.xml.base import Serializable, SerializableArray, create_text_node
from sarpy.io.xml.descriptors import FloatDescriptor, IntegerDescriptor, \
    StringDescriptor, StringEnumDescriptor, \
    SerializableDescriptor, SerializableArrayDescriptor, UnitVectorDescriptor
from sarpy.io.complex.sicd_elements.base import SerializableCPArray, SerializableCPArrayDescriptor
from sarpy.io.complex.sicd_elements.blocks import XYZType, LatLonType, LatLonCornerType
from sarpy.io.complex.sicd_elements.GeoData import SCPType
from .base import DEFAULT_STRICT
from .blocks import AreaType, LSType, LSVertexType


class IARPType(SCPType):
    """
    The Image Area Reference Point (IARP), which is the origin of the Image Area Coordinate system.
    Note that setting one of ECF or LLH will implicitly set the other to its corresponding matched value.
    """


class ECFPlanarType(Serializable):
    """
    Parameters for a planar surface defined in ECF coordinates. The reference
    surface is a plane that contains the IARP.
    """

    _fields = ('uIAX', 'uIAY')
    _required = _fields
    # descriptors
    uIAX = UnitVectorDescriptor(
        'uIAX', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='Image Area X-coordinate (IAX) unit vector in ECF coordinates. '
                  'For stripmap collections, uIAX ALWAYS points in the direction '
                  'of the scanning footprint.')  # type: XYZType
    uIAY = UnitVectorDescriptor(
        'uIAY', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='Image Area Y-coordinate (IAY) unit vector in ECF '
                  'coordinates. This should be perpendicular to '
                  'uIAX.')  # type: XYZType

    def __init__(self, uIAX=None, uIAY=None, **kwargs):
        """

        Parameters
        ----------
        uIAX : XYZType|numpy.ndarray|list|tuple
        uIAY : XYZType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.uIAX = uIAX
        self.uIAY = uIAY
        super(ECFPlanarType, self).__init__(**kwargs)


class LLPlanarType(Serializable):
    """
    Parameters for Lat/Lon planar surface of constant HAE, implicitly assumed to be
    the HAE at the `IARP`.
    """

    _fields = ('uIAXLL', 'uIAYLL')
    _required = _fields
    # descriptors
    uIAXLL = SerializableDescriptor(
        'uIAXLL', LatLonType, _required, strict=DEFAULT_STRICT,
        docstring='Image coordinate IAX *"unit vector"* expressed as an increment '
                  'in latitude and longitude corresponding to a 1.0 meter increment '
                  'in image coordinate `IAX`.')  # type: LatLonType
    uIAYLL = SerializableDescriptor(
        'uIAYLL', LatLonType, _required, strict=DEFAULT_STRICT,
        docstring='Image coordinate IAY *"unit vector"* expressed as an increment '
                  'in latitude and longitude corresponding to a 1.0 meter increment '
                  'in image coordinate `IAY`.')  # type: LatLonType

    def __init__(self, uIAXLL=None, uIAYLL=None, **kwargs):
        """

        Parameters
        ----------
        uIAXLL : LatLonType|numpy.ndarray|list|tuple
        uIAYLL : LatLonType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.uIAXLL = uIAXLL
        self.uIAYLL = uIAYLL
        super(LLPlanarType, self).__init__(**kwargs)


class ReferenceSurfaceType(Serializable):
    """
    Parameters that define the Reference Surface used for the product.
    """
    _fields = ('Planar', 'HAE')
    _required = ()
    _choice = ({'required': True, 'collection': _fields}, )
    # descriptors
    Planar = SerializableDescriptor(
        'Planar', ECFPlanarType, _required, strict=DEFAULT_STRICT,
        docstring='The ECF planar surface definition.')  # type: Union[None, ECFPlanarType]
    HAE = SerializableDescriptor(
        'HAE', LLPlanarType, _required, strict=DEFAULT_STRICT,
        docstring='The HAE surface definition.')  # type: Union[None, LLPlanarType]

    def __init__(self, Planar=None, HAE=None, **kwargs):
        """

        Parameters
        ----------
        Planar : ECFPlanarType|None
        HAE : LLPlanarType|None
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Planar = Planar
        self.HAE = HAE
        super(ReferenceSurfaceType, self).__init__(**kwargs)


###########
# Image grid definition

class IAXExtentType(Serializable):
    """
    Increasing line index is in the +IAX direction.
    """
    _fields = ('LineSpacing', 'FirstLine', 'NumLines')
    _required = _fields
    _numeric_format = {'LineSpacing': '0.17G'}
    # descriptors
    LineSpacing = FloatDescriptor(
        'LineSpacing', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='The line spacing, in meters.')  # type: float
    FirstLine = IntegerDescriptor(
        'FirstLine', _required, strict=DEFAULT_STRICT,
        docstring='Index of the first line.')  # type: int
    NumLines = IntegerDescriptor(
        'NumLines', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='Number of lines.')  # type: int

    def __init__(self, LineSpacing=None, FirstLine=None, NumLines=None, **kwargs):
        """

        Parameters
        ----------
        LineSpacing : float
        FirstLine : int
        NumLines : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.LineSpacing = LineSpacing
        self.FirstLine = FirstLine
        self.NumLines = NumLines
        super(IAXExtentType, self).__init__(**kwargs)


class IAYExtentType(Serializable):
    """
    Increasing sample index is in the +IAY direction.
    """

    _fields = ('SampleSpacing', 'FirstSample', 'NumSamples')
    _required = _fields
    _numeric_format = {'SampleSpacing': '0.17G'}
    # descriptors
    SampleSpacing = FloatDescriptor(
        'SampleSpacing', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Sample spacing, in meters.')  # type: float
    FirstSample = IntegerDescriptor(
        'FirstSample', _required, strict=DEFAULT_STRICT,
        docstring='Index of the first sample.')  # type: int
    NumSamples = IntegerDescriptor(
        'NumSamples', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='Number of samples.')  # type: int

    def __init__(self, SampleSpacing=None, FirstSample=None, NumSamples=None, **kwargs):
        """

        Parameters
        ----------
        SampleSpacing : float
        FirstSample : int
        NumSamples : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.SampleSpacing = SampleSpacing
        self.FirstSample = FirstSample
        self.NumSamples = NumSamples
        super(IAYExtentType, self).__init__(**kwargs)


class SegmentType(Serializable):
    """
    Rectangle segment.
    """

    _fields = ('Identifier', 'StartLine', 'StartSample', 'EndLine', 'EndSample', 'SegmentPolygon')
    _required = ('Identifier', 'StartLine', 'StartSample', 'EndLine', 'EndSample')
    _collections_tags = {'SegmentPolygon': {'array': True, 'child_tag': 'SV'}}
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='String that uniquely identifies the Image Segment.')  # type: str
    StartLine = IntegerDescriptor(
        'StartLine', _required, strict=DEFAULT_STRICT,
        docstring='Start line of the segment.')  # type: int
    StartSample = IntegerDescriptor(
        'StartSample', _required, strict=DEFAULT_STRICT,
        docstring='Start sample of the segment.')  # type: int
    EndLine = IntegerDescriptor(
        'EndLine', _required, strict=DEFAULT_STRICT,
        docstring='End line of the segment.')  # type: int
    EndSample = IntegerDescriptor(
        'EndSample', _required, strict=DEFAULT_STRICT,
        docstring='End sample of the segment.')  # type: int
    SegmentPolygon = SerializableArrayDescriptor(
        'SegmentPolygon', LSVertexType, _collections_tags, _required,
        strict=DEFAULT_STRICT, minimum_length=3,
        docstring='Polygon that describes a portion of the segment '
                  'rectangle.')  # type: Union[SerializableArray, List[LSVertexType]]

    def __init__(self, Identifier=None, StartLine=None, StartSample=None, EndLine=None,
                 EndSample=None, SegmentPolygon=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        StartLine : int
        StartSample : int
        EndLine : int
        EndSample : int
        SegmentPolygon : SerializableArray|List[LSVertexType]|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.StartLine = StartLine
        self.StartSample = StartSample
        self.EndLine = EndLine
        self.EndSample = EndSample
        self.SegmentPolygon = SegmentPolygon
        super(SegmentType, self).__init__(**kwargs)


class SegmentListType(SerializableArray):
    _set_size = False
    _set_index = False

    @property
    def NumSegments(self):
        return self.size

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT):
        anode = super(SegmentListType, self).to_node(
            doc, tag, ns_key=ns_key, parent=parent, check_validity=check_validity, strict=strict)
        create_text_node(
            doc, 'NumSegments' if ns_key is None else '{}:NumSegments'.format(ns_key),
            '{0:d}'.format(self.NumSegments), parent=anode)
        return anode


class ImageGridType(Serializable):
    """
    Parameters that describe a geo-referenced image grid for image data products
    that may be formed from the CPHD signal array(s).
    """

    _fields = ('Identifier', 'IARPLocation', 'IAXExtent', 'IAYExtent', 'SegmentList')
    _required = ('IARPLocation', 'IAXExtent', 'IAYExtent')
    _collections_tags = {'SegmentList': {'array': True, 'child_tag': 'Segment'}}
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='String that uniquely identifies the Image Grid.')  # type: Union[None, str]
    IARPLocation = SerializableDescriptor(
        'IARPLocation', LSType, _required, strict=DEFAULT_STRICT,
        docstring='IARP grid location. Grid locations indexed by (line, sample) or (L,S). '
                  'Image grid line and sample are pixel-centered indices.')  # type: LSType
    IAXExtent = SerializableDescriptor(
        'IAXExtent', IAXExtentType, _required, strict=DEFAULT_STRICT,
        docstring='Increasing line index is in the +IAX direction.')  # type: IAXExtentType
    IAYExtent = SerializableDescriptor(
        'IAYExtent', IAYExtentType, _required, strict=DEFAULT_STRICT,
        docstring='Increasing sample index is in the +IAY direction.')  # type: IAYExtentType
    SegmentList = SerializableArrayDescriptor(
        'SegmentList', SegmentType, _collections_tags, _required, strict=DEFAULT_STRICT,
        array_extension=SegmentListType,
        docstring='List of image grid segments defined relative to the image '
                  'grid.')  # type: Union[SegmentListType, List[SegmentType]]

    def __init__(self, Identifier=None, IARPLocation=None, IAXExtent=None, IAYExtent=None,
                 SegmentList=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : None|str
        IARPLocation : LSType
        IAXExtent : IAXExtentType
        IAYExtent : IAYExtentType
        SegmentList : SegmentListType|List[SegmentType]|numpy.array|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.IARPLocation = IARPLocation
        self.IAXExtent = IAXExtent
        self.IAYExtent = IAYExtent
        self.SegmentList = SegmentList
        super(ImageGridType, self).__init__(**kwargs)


###############

class SceneCoordinatesType(Serializable):
    """
    Parameters that define geographic coordinates for in the imaged scene.
    """

    _fields = (
        'EarthModel', 'IARP', 'ReferenceSurface', 'ImageArea', 'ImageAreaCornerPoints',
        'ExtendedArea', 'ImageGrid')
    _required = ('EarthModel', 'IARP', 'ReferenceSurface', 'ImageArea', 'ImageAreaCornerPoints')
    _collections_tags = {
        'ImageAreaCornerPoints': {'array': True, 'child_tag': 'IACP'}}
    # descriptors
    EarthModel = StringEnumDescriptor(
        'EarthModel', ('WGS_84', ), _required, strict=DEFAULT_STRICT, default_value='WGS_84',
        docstring='Specifies the earth model used for specifying geodetic coordinates. All heights are '
                  'Height Above the Ellipsoid (HAE) unless specifically '
                  'noted.')  # type: str
    IARP = SerializableDescriptor(
        'IARP', IARPType, _required, strict=DEFAULT_STRICT,
        docstring='Image Area Reference Point (IARP). The IARP is the origin of '
                  'the Image Area Coordinate system.')  # type: IARPType
    ReferenceSurface = SerializableDescriptor(
        'ReferenceSurface', ReferenceSurfaceType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that define the Reference Surface used for the '
                  'product.')  # type: ReferenceSurfaceType
    ImageArea = SerializableDescriptor(
        'ImageArea', AreaType, _required, strict=DEFAULT_STRICT,
        docstring='Image Area is defined by a rectangle aligned with Image Area coordinates (IAX, IAY). '
                  'May be reduced by the optional polygon.')  # type: AreaType
    ImageAreaCornerPoints = SerializableCPArrayDescriptor(
        'ImageAreaCornerPoints', LatLonCornerType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Image Area Corner Points (IACPs) that bound the full resolution '
                  'image area.')  # type: Union[SerializableCPArray, List[LatLonCornerType]]
    ExtendedArea = SerializableDescriptor(
        'ExtendedArea', AreaType, _required, strict=DEFAULT_STRICT,
        docstring='Extended Area is defined by a rectangle aligned with Image Area coordinates '
                  '(IAX, IAY). May be reduced by the optional polygon.')  # type: Union[None, AreaType]
    ImageGrid = SerializableDescriptor(
        'ImageGrid', ImageGridType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe a geo-referenced image grid for image data '
                  'products that may be formed from the CPHD signal '
                  'array(s).')  # type: ImageGridType

    def __init__(self, EarthModel='WGS_84', IARP=None, ReferenceSurface=None,
                 ImageArea=None, ImageAreaCornerPoints=None, ExtendedArea=None,
                 ImageGrid=None, **kwargs):
        """

        Parameters
        ----------
        EarthModel : None|str
        IARP : IARPType
        ReferenceSurface : ReferenceSurfaceType
        ImageArea : AreaType
        ImageAreaCornerPoints : SerializableCPArray|List[LatLonCornerType]|numpy.ndarray|list|tuple
        ExtendedArea : None|AreaType
        ImageGrid : None|ImageGridType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.EarthModel = EarthModel
        self.IARP = IARP
        self.ReferenceSurface = ReferenceSurface
        self.ImageArea = ImageArea
        self.ImageAreaCornerPoints = ImageAreaCornerPoints
        self.ExtendedArea = ExtendedArea
        self.ImageGrid = ImageGrid
        super(SceneCoordinatesType, self).__init__(**kwargs)




__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from ..tre_elements import TREExtension, TREElement


logger = logging.getLogger(__name__)


def get_ref_type_length(typ_val):
    typ_val = typ_val.lower()
    if typ_val in ['06a', '06c']:
        return 11
    elif typ_val == '06b':
        return 12
    elif typ_val in ['06d', '06e', '06f']:
        return 8
    elif typ_val in [
            '07b', '07d', '07h', '08a', '08b', '08c', '08d', '08e', '08f', '08g',
            '08h', '08i', '09a', '09b', '09c', '09d']:
        return 10
    elif typ_val in ['07c', '07f', '07g', '10a', '10b', '10c']:
        return 9
    else:
        logger.error(
            'An unknown type value {} was found when deserializing a SENSRB TRE object.\n\t'
            'Something may fail in this deserialization.')
        return None


class POINT(TREElement):
    def __init__(self, value):
        super(POINT, self).__init__()
        self.add_field('P_ROW', 's', 8, value)
        self.add_field('P_COLUMN', 's', 8, value)
        self.add_field('P_LATITUDE', 's', 10, value)
        self.add_field('P_LONGITUDE', 's', 11, value)
        self.add_field('P_ELEVATION', 's', 6, value)
        self.add_field('P_RANGE', 's', 8, value)


class POINT_SET(TREElement):
    def __init__(self, value):
        super(POINT_SET, self).__init__()
        self.add_field('POINT_SET_TYPE', 's', 25, value)
        self.add_field('POINT_COUNT', 'd', 3, value)
        self.add_loop('POINTs', self.POINT_COUNT, POINT, value)


class TIME_STAMP(TREElement):
    def __init__(self, value, time_len):
        super(TIME_STAMP, self).__init__()
        self.add_field('TIME_STAMP_TIME', 's', 12, value)
        if time_len is not None:
            self.add_field('TIME_STAMP_VALUE', 's', time_len, value)


class TIME_STAMPED_DATA(TREElement):
    def __init__(self, value):
        super(TIME_STAMPED_DATA, self).__init__()
        self.add_field('TIME_STAMP_TYPE', 's', 3, value)
        self.add_field('TIME_STAMP_COUNT', 'd', 4, value)
        self.add_loop(
            'TIME_STAMPs', self.TIME_STAMP_COUNT, TIME_STAMP, value,
            get_ref_type_length(self.TIME_STAMP_TYPE))


class PIXEL_REFERENCE(TREElement):
    def __init__(self, value, pixel_ref_len):
        super(PIXEL_REFERENCE, self).__init__()
        self.add_field('PIXEL_REFERENCE_ROW', 's', 8, value)
        self.add_field('PIXEL_REFERENCE_COLUMN', 's', 8, value)
        if pixel_ref_len is not None:
            self.add_field('PIXEL_REFERENCE_VALUE', 's', pixel_ref_len, value)


class PIXEL_REFERENCED_DATA(TREElement):
    def __init__(self, value):
        super(PIXEL_REFERENCED_DATA, self).__init__()
        self.add_field('PIXEL_REFERENCE_TYPE', 's', 3, value)
        self.add_field('PIXEL_REFERENCE_COUNT', 'd', 4, value)
        self.add_loop(
            'PIXEL_REFERENCEs', self.PIXEL_REFERENCE_COUNT, PIXEL_REFERENCE, value,
            get_ref_type_length(self.PIXEL_REFERENCE_TYPE))


class UNCERTAINTY(TREElement):
    def __init__(self, value):
        super(UNCERTAINTY, self).__init__()
        self.add_field('UNCERTAINTY_FIRST_TYPE', 's', 11, value)
        self.add_field('UNCERTAINTY_SECOND_TYPE', 's', 11, value)
        self.add_field('UNCERTAINTY_VALUE', 's', 10, value)


class PARAMETER(TREElement):
    def __init__(self, value):
        super(PARAMETER, self).__init__()
        self.add_field('PARAMETER_VALUE', 'b', self.PARAMETER_SIZE, value)


class ADDITIONAL_PARAMETER(TREElement):
    def __init__(self, value):
        super(ADDITIONAL_PARAMETER, self).__init__()
        self.add_field('PARAMETER_NAME', 's', 25, value)
        self.add_field('PARAMETER_SIZE', 's', 3, value)
        self.add_field('PARAMETER_COUNT', 'd', 4, value)
        self.add_loop('PARAMETERs', self.PARAMETER_COUNT, PARAMETER, value)


class SENSRBType(TREElement):
    def __init__(self, value):
        super(SENSRBType, self).__init__()

        self.add_field('GENERAL_DATA', 's', 1, value)
        if self.GENERAL_DATA == 'Y':
            self.add_field('SENSOR', 's', 25, value)
            self.add_field('SENSOR_URI', 's', 32, value)
            self.add_field('PLATFORM', 's', 25, value)
            self.add_field('PLATFORM_URI', 's', 32, value)
            self.add_field('OPERATION_DOMAIN', 's', 10, value)
            self.add_field('CONTENT_LEVEL', 's', 1, value)
            self.add_field('GEODETIC_SYSTEM', 's', 5, value)
            self.add_field('GEODETIC_TYPE', 's', 1, value)
            self.add_field('ELEVATION_DATUM', 's', 3, value)
            self.add_field('LENGTH_UNIT', 's', 2, value)
            self.add_field('ANGULAR_UNIT', 's', 3, value)
            self.add_field('START_DATE', 's', 8, value)
            self.add_field('START_TIME', 's', 14, value)
            self.add_field('END_DATE', 's', 8, value)
            self.add_field('END_TIME', 's', 14, value)
            self.add_field('GENERATION_COUNT', 's', 2, value)
            self.add_field('GENERATION_DATE', 's', 8, value)
            self.add_field('GENERATION_TIME', 's', 10, value)

        self.add_field('SENSOR_ARRAY_DATA', 's', 1, value)
        if self.SENSOR_ARRAY_DATA == 'Y':
            self.add_field('DETECTION', 's', 20, value)
            self.add_field('ROW_DETECTORS', 's', 8, value)
            self.add_field('COLUMN_DETECTORS', 's', 8, value)
            self.add_field('ROW_METRIC', 's', 8, value)
            self.add_field('COLUMN_METRIC', 's', 8, value)
            self.add_field('FOCAL_LENGTH', 's', 8, value)
            self.add_field('ROW_FOV', 's', 8, value)
            self.add_field('COLUMN_FOV', 's', 8, value)
            self.add_field('CALIBRATED', 's', 1, value)

        self.add_field('SENSOR_CALIBRATION_DATA', 's', 1, value)
        if self.SENSOR_CALIBRATION_DATA == 'Y':
            self.add_field('CALIBRATION_UNIT', 's', 2, value)
            self.add_field('PRINCIPAL_POINT_OFFSET_X', 's', 9, value)
            self.add_field('PRINCIPAL_POINT_OFFSET_Y', 's', 9, value)
            self.add_field('RADIAL_DISTORT_1', 's', 12, value)
            self.add_field('RADIAL_DISTORT_2', 's', 12, value)
            self.add_field('RADIAL_DISTORT_3', 's', 12, value)
            self.add_field('RADIAL_DISTORT_LIMIT', 's', 9, value)
            self.add_field('DECENT_DISTORT_1', 's', 12, value)
            self.add_field('DECENT_DISTORT_2', 's', 12, value)
            self.add_field('AFFINITY_DISTORT_1', 's', 12, value)
            self.add_field('AFFINITY_DISTORT_2', 's', 12, value)
            self.add_field('CALIBRATION_DATE', 's', 8, value)

        self.add_field('IMAGE_FORMATION_DATA', 's', 1, value)
        if self.IMAGE_FORMATION_DATA == 'Y':
            self.add_field('METHOD', 's', 15, value)
            self.add_field('MODE', 's', 3, value)
            self.add_field('ROW_COUNT', 's', 8, value)
            self.add_field('COLUMN_COUNT', 's', 8, value)
            self.add_field('ROW_SET', 's', 8, value)
            self.add_field('COLUMN_SET', 's', 8, value)
            self.add_field('ROW_RATE', 's', 10, value)
            self.add_field('COLUMN_RATE', 's', 10, value)
            self.add_field('FIRST_PIXEL_ROW', 's', 8, value)
            self.add_field('FIRST_PIXEL_COLUMN', 's', 8, value)
            self.add_field('TRANSFORM_PARAMS', 'd', 1, value)
            for i in range(self.TRANSFORM_PARAMS):
                self.add_field('TRANSFORM_PARAM_{}'.format(i+1), 's', 12, value)
        self.add_field('REFERENCE_TIME', 's', 12, value)
        self.add_field('REFERENCE_ROW', 's', 8, value)
        self.add_field('REFERENCE_COLUMN', 's', 8, value)
        self.add_field('LATITUDE_OR_X', 's', 11, value)
        self.add_field('LONGITUDE_OR_Y', 's', 12, value)
        self.add_field('ALTITUDE_OR_Z', 's', 11, value)
        self.add_field('SENSOR_X_OFFSET', 's', 8, value)
        self.add_field('SENSOR_Y_OFFSET', 's', 8, value)
        self.add_field('SENSOR_Z_OFFSET', 's', 8, value)

        self.add_field('ATTITUDE_EULER_ANGLES', 's', 1, value)
        if self.ATTITUDE_EULER_ANGLES == 'Y':
            self.add_field('SENSOR_ANGLE_MODEL', 's', 1, value)
            self.add_field('SENSOR_ANGLE_1', 's', 10, value)
            self.add_field('SENSOR_ANGLE_2', 's', 9, value)
            self.add_field('SENSOR_ANGLE_3', 's', 10, value)
            self.add_field('PLATFORM_RELATIVE', 's', 1, value)
            self.add_field('PLATFORM_HEADING', 's', 9, value)
            self.add_field('PLATFORM_PITCH', 's', 9, value)
            self.add_field('PLATFORM_ROLL', 's', 10, value)

        self.add_field('ATTITUDE_UNIT_VECTORS', 's', 1, value)
        if self.ATTITUDE_UNIT_VECTORS == 'Y':
            self.add_field('ICX_NORTH_OR_X', 's', 10, value)
            self.add_field('ICX_EAST_OR_Y', 's', 10, value)
            self.add_field('ICX_DOWN_OR_Z', 's', 10, value)
            self.add_field('ICY_NORTH_OR_X', 's', 10, value)
            self.add_field('ICY_EAST_OR_Y', 's', 10, value)
            self.add_field('ICY_DOWN_OR_Z', 's', 10, value)
            self.add_field('ICZ_NORTH_OR_X', 's', 10, value)
            self.add_field('ICZ_EAST_OR_Y', 's', 10, value)
            self.add_field('ICZ_DOWN_OR_Z', 's', 10, value)

        self.add_field('ATTITUDE_QUATERNION', 's', 1, value)
        if self.ATTITUDE_QUATERNION == 'Y':
            self.add_field('ATTITUDE_Q1', 's', 10, value)
            self.add_field('ATTITUDE_Q2', 's', 10, value)
            self.add_field('ATTITUDE_Q3', 's', 10, value)
            self.add_field('ATTITUDE_Q4', 's', 10, value)

        self.add_field('SENSOR_VELOCITY_DATA', 's', 1, value)
        if self.SENSOR_VELOCITY_DATA == 'Y':
            self.add_field('VELOCITY_NORTH_OR_X', 's', 9, value)
            self.add_field('VELOCITY_EAST_OR_Y', 's', 9, value)
            self.add_field('VELOCITY_DOWN_OR_Z', 's', 9, value)

        self.add_field('POINT_SET_DATA', 'd', 2, value)
        self.add_loop('POINT_SETs', self.POINT_SET_DATA, POINT_SET, value)

        self.add_field('TIME_STAMPED_DATA_SETS', 'd', 2, value)
        self.add_loop('TIME_STAMPED_DATAs', self.TIME_STAMPED_DATA_SETS, TIME_STAMPED_DATA, value)

        self.add_field('PIXEL_REFERENCED_DATA_SETS', 'd', 2, value)
        self.add_loop('PIXEL_REFERENCED_DATAs', self.PIXEL_REFERENCED_DATA_SETS, PIXEL_REFERENCED_DATA, value)

        self.add_field('UNCERTAINTY_DATA', 'd', 3, value)
        self.add_loop('UNCERTAINTYs', self.UNCERTAINTY_DATA, UNCERTAINTY, value)

        self.add_field('ADDITIONAL_PARAMETER_DATA', 'd', 3, value)
        self.add_loop('ADDITIONAL_PARAMETERs', self.ADDITIONAL_PARAMETER_DATA, ADDITIONAL_PARAMETER, value)


class SENSRB(TREExtension):
    _tag_value = 'SENSRB'
    _data_type = SENSRBType
